"""

Region module for geopyv.

"""
import logging
import numpy as np
import geopyv as gp
from geopyv.object import Object
import re

log = logging.getLogger(__name__)


class RegionBase(Object):
    """

    Region base class.

    """

    def __init__(self):
        super().__init__(object_type="geometry.Region")
        """

        Mesh base class initialiser.

        """

    def _report(self, msg, error_type):
        if msg and error_type != "Warning":
            log.error(msg)
        elif msg and error_type == "Warning":
            log.warning(msg)
            return True
        if error_type == "ValueError" and msg:
            raise ValueError(msg)
        elif error_type == "TypeError" and msg:
            raise TypeError(msg)
        elif error_type == "IndexError" and msg:
            raise IndexError(msg)


class Region(RegionBase):
    def __init__(
        self,
        *,
        shape=None,
        centre=None,
        nodes=None,
        option="F",
        hard=True,
        compensate=True,
    ):
        """
        hard : bool, optional
            Boolean to control whether the nodes is included in the
            binary mask. True -included, False - not included.
            Defaults to True.
        option :
            "D" : Defined: a controlled region.
            "S" : Static: an untracked region.
            "R" : Rigid: a tracked region, unable to deform.
            "F" : Flexible: a tracked region, able to deform.
        """
        self._shape = shape
        self._option = option
        if self._option == "D":
            centres = [np.mean(nodes[i], axis=0) for i in range(np.shape(nodes)[0])]
            self._nodes = nodes[0]
            self._centre = centres[0]
        else:
            self._nodes = nodes
            self._centre = centre
            nodes = [nodes]
            centres = [centre]
        self._hard = hard
        self._compensate = compensate

        self.solved = False

        self.data = {
            "type": "geometry.Region",
            "solved": self.solved,
            "shape": self._shape,
            "hard": self._hard,
            "compensate": self._compensate,
            "specifics": self._specifics,
            "centres": centres,
            "nodes": nodes,
            "counter": 0,
            "ref_index": None,
        }
        self._reference_update_register = []

    def _store(self, warp):
        if self._option == "R":
            local_coordinates = self._nodes - self._centre
            theta = (warp[3] - warp[4]) / 2
            rot = np.asarray(
                [[np.cos(theta), np.sin(theta)], [-np.sin(theta), np.cos(theta)]]
            )
            self.data["nodes"].append(self._centre + warp[:2] + local_coordinates @ rot)
            self.data["centres"].append(self._centre + warp[:2])
        elif self._option == "F":
            self.data["nodes"].append(self._nodes + warp)
            self.data["centres"].append(self._centre + np.mean(warp, axis=0))
        self._solved = True
        self.data["solved"] = self._solved
        self.data["counter"] += 1

    def _update(self, f_img_filepath):
        if self._option != "S":
            f_index = int(
                re.findall(
                    r"\d+",
                    f_img_filepath,
                )[-1]
            )
            if f_index != self.data["ref_index"]:
                self.data["ref_index"] = f_index
                self._reference_update_register.append(f_index)
                self._nodes = self.data["nodes"][self.data["counter"]]
                self._centre = self.data["centres"][self.data["counter"]]


class Circle(Region):
    """

    Circular Region template class.

    """

    def __init__(
        self,
        *,
        centre=None,
        radius=50.0,
        size=20.0,
        option="F",
        hard=True,
        compensate=True,
    ):
        """

        Class for circular Region. Subclassed from Region.

        Parameters
        ----------
        centre : numpy.ndarray(2,)
            Central coordinate of the circular Region.

        Attributes
        ----------

        """
        # Input check.
        if self._report(
            gp.check._check_type(centre, "centre", [np.ndarray]), "Warning"
        ):
            image = gp.io._load_f_img()
            selector = gp.gui.selectors.coordinate.coordinateSelector()
            centre = selector.select(image, gp.templates.Circle(50))
        check = gp.check._check_type(radius, "radius", [float])
        if check:
            try:
                radius = float(radius)
                self._report(gp.check._conversion(radius, "radius", float), "Warning")
            except Exception:
                self._report(check, "TypeError")
        self._report(
            gp.check._check_range(radius, "radius", 0),
            "ValueError",
        )
        check = gp.check._check_type(size, "size", [float])
        if check:
            try:
                size = float(size)
                self._report(gp.check._conversion(size, "size", float), "Warning")
            except Exception:
                self._report(check, "TypeError")
        self._report(
            gp.check._check_range(size, "size", 0),
            "ValueError",
        )
        self._report(gp.check._check_type(hard, "hard", [bool]), "TypeError")

        # Store unique variables.
        self._size = size
        self._radius = radius
        self._specifics = {
            "size": self._size,
            "radius": self._radius,
        }

        # Defining nodes.
        number_points = np.maximum(
            6, int(2 * np.pi * radius / size)
        )  # Use a minimum of six points irrespective of target size defined.
        theta = np.linspace(0, 2 * np.pi, number_points, endpoint=False)
        x = radius * np.cos(theta) + centre[0]
        y = radius * np.sin(theta) + centre[1]
        nodes = np.column_stack((x, y))

        # Store general variables.
        super().__init__(
            shape="Circle",
            centre=centre,
            nodes=nodes,
            option=option,
            hard=hard,
            compensate=compensate,
        )


class Path(Region):
    """

    Path Region template class.

    """

    def __init__(
        self,
        *,
        centre=None,
        nodes=None,
        option="F",
        hard=True,
        compensate=True,
        radius=25,
    ):
        """

        Class for circular Region. Subclassed from Region.

        Parameters
        ----------
        centre : numpy.ndarray(2,)
            Central coordinate of the circular Region.

        Attributes
        ---------

        """
        # Input check.
        if self._report(
            gp.check._check_type(centre, "centre", [np.ndarray, type(None)]), "Warning"
        ):
            image = gp.io._load_f_img()
            selector = gp.gui.selectors.coordinate.coordinateSelector()
            centre = selector.select(image, gp.templates.Circle(50))
        check = gp.check._check_type(nodes, "nodes", [np.ndarray])
        if check:
            try:
                nodes = np.asarray(nodes)
                self._report(
                    gp.check._conversion(nodes, "nodes", np.ndarray, False),
                    "Warning",
                )
            except Exception:
                self._report(check, "TypeError")
        # self._report(gp.check._check_dim(nodes, "nodes", 2), "ValueError")
        # self._report(gp.check._check_axis(nodes, "nodes", 1, [2]), "ValueError")

        if centre is None:
            centre = np.mean(nodes, axis=0)

        # Store unique variables.
        self._specifics = {"radius": radius}

        # Store general variables.
        super().__init__(
            shape="Path",
            centre=centre,
            nodes=nodes,
            option=option,
            hard=hard,
            compensate=compensate,
        )


class RegionResults(RegionBase):
    """
    RegionResults class for geopyv.

    Parameters
    ----------
    data : dict
        geopyv data dict from Region object.

    Attributes
    ----------
    data : dict
        geopyv data dict from Region object.

    """

    def __init__(self, data):
        """Initialisation of geopyv RegionResults class."""
        self.data = data



import logging
import numpy as np
import matplotlib.pyplot as plt
import geopyv.templates as templates

log = logging.getLogger(__name__)


class CoordinateSelector:
    def select(self, f_img, template):
        """Method to select f_coord if not supplied by the user."""
        log.warning(
            "No coordinate supplied. Please select the target "
            "coordinate for the subset."
        )
        self.f_img = f_img
        self.template = template
        self.x = None
        self.y = None
        self.f, self.ax = plt.subplots(
            num="Right click on the target coordinate for the subset "
            "and close to save"
        )
        self.f.canvas.mpl_connect("button_press_event", self.on_click)
        self.f.canvas.mpl_connect("close_event", self.selected)
        self.ax.imshow(self.f_img.image_gs, cmap="gist_gray")
        self.ax.set_axis_off()
        self.point = None
        self.boundary = None
        plt.tight_layout()
        plt.show()
        self.coord = np.asarray([self.x, self.y])
        return self.coord

    def on_click(self, event):
        """

        Method to store and plot the currently selected coordinate in self.f_coord.

        """
        if event.button == 3:
            if event.xdata is not None and event.ydata is not None:
                if (
                    event.xdata > self.template.size
                    and event.xdata
                    < np.shape(self.f_img.image_gs)[0] - self.template.size
                ):
                    if (
                        event.ydata > self.template.size
                        and event.ydata
                        < np.shape(self.f_img.image_gs)[1] - self.template.size
                    ):
                        self.x = np.round(event.xdata, 0)
                        self.y = np.round(event.ydata, 0)
                        self.f_coord = np.asarray([self.x, self.y])
                        if self.point is not None:
                            self.point[0].remove()
                        if self.boundary is not None:
                            self.boundary[0].remove()
                        self.point = self.ax.plot(
                            self.x,
                            self.y,
                            marker="+",
                            color="b",
                            zorder=10,
                        )
                        if type(self.template) == templates.Circle:
                            theta = np.linspace(0, 2 * np.pi, 150)
                            radius = self.template.size
                            x = self.x + radius * np.cos(theta)
                            y = self.y + radius * np.sin(theta)
                            self.boundary = self.ax.plot(x, y, color="b")
                        elif type(self.template) == templates.Square:
                            x = [
                                self.x - self.template.size,
                                self.x - self.template.size,
                                self.x + self.template.size,
                                self.x + self.template.size,
                                self.x - self.template.size,
                            ]
                            y = [
                                self.y - self.template.size,
                                self.y + self.template.size,
                                self.y + self.template.size,
                                self.y - self.template.size,
                                self.y - self.template.size,
                            ]
                            self.boundary = self.ax.plot(x, y, color="b")
                        self.f.canvas.draw()
                        self.f.canvas.flush_events()

    def selected(self, event):
        """Method to print the selected coordinates."""
        log.info("Coordinate selected: {}, {}".format(self.x, self.y))



import geopyv.check
import geopyv.calibration
import geopyv.log
import geopyv.geometry
import geopyv.gui
import geopyv.image
import geopyv._image_extensions
import geopyv.io
import geopyv.field
import geopyv.mesh
import geopyv.object
import geopyv.particle
import geopyv.plots
import geopyv.sequence
import geopyv.speckle
import geopyv.subset
import geopyv._subset_extensions
import geopyv.templates
import geopyv.validation
import logging

# Initialise log at default settings.
level = logging.INFO
geopyv.log.initialise(level)
log = logging.getLogger(__name__)
log.debug("Initialised geopyv log.")



from PySide6.QtWidgets import QApplication, QFileDialog, QWidget


class ImageSelector:
    def __init__(self):
        """

        Graphical user interface to allow the user to select an image
        using the native file browser on the host OS.

        """
        app = QApplication.instance()
        if app is None:
            app = QApplication()
        self.widget = QWidget()

    def get_path(self, message, directory):
        """

        Method to get the path of the selected file.

        Parameters
        ----------
        message : str
            Message to display on the file browser window header.
        directory : str
            Directory to start from.

        Returns
        -------
        path : str
            Path to selected image file.

        """
        path, _ = QFileDialog.getOpenFileName(
            self.widget,
            directory,
            message,
            "Image files (*.png *.bmp *.jpg)",
        )

        return path



"""

IO module for geopyv.

"""
import importlib
import logging
import pickle
import os
import geopyv as gp
from alive_progress import alive_bar

log = logging.getLogger(__name__)


def load(*, filename=None, directory=None, old_format=False, verbose=True):
    """

    Function to load a geopyv data object into the workspace. If no filename
     is provided, the host OS default file browser will be used to allow the
      user to select a geopyv data file with .pyv extension.


    Parameters
    ----------
    filename : str, optional
        Name of file containing a geopy data object.


    Returns
    -------
    object : geopyv.object.Object
        The geopyv data object loaded.


    .. note::
        * Any .pyv object can be loaded with this function.
        * The data object will be loaded into a `ObjectResults` instance where
          `Object` represents the instance type that generated the data. For example,
          data from a `Subset` instance will be loaded into a `SubsetResults` instance.

    """
    cwd = os.getcwd()
    if directory is None and cwd not in filename:
        directory = os.getcwd()
    if filename is None:
        dialog = gp.gui.selectors.file.FileSelector()
        filepath = dialog.get_path("Select geopyv data file", directory)
    else:
        ext = ".pyv"
        # filepath = filename + ext
        if ext in filename:
            filepath = directory + "/" + filename
        else:
            filepath = directory + "/" + filename + ext
    try:
        with open(filepath, "rb") as infile:
            if verbose:
                message = "Loading geopyv object"
                with alive_bar(dual_line=True, bar=None, title=message) as bar:
                    bar.text = "-> Loading object from {filepath}...".format(
                        filepath=filepath
                    )
                    data = pickle.load(infile)
                    bar()
            else:
                data = pickle.load(infile)
            object_type = data["type"]
            if verbose:
                log.info(
                    "Loaded {object_type} object from {filepath}.".format(
                        object_type=object_type, filepath=filepath
                    )
                )
            if "." in data["type"]:
                class_name = data["type"].split(".")[-1] + "Results"
            else:
                class_name = data["type"] + "Results"
            module = importlib.import_module("geopyv." + object_type.lower())
            results_instance = getattr(module, class_name)
            return results_instance(data)
    except Exception:
        log.error(
            "File does not exist at the path supplied:\n{filepath}".format(
                filepath=filepath
            )
        )
        raise FileExistsError


def save(*, object, directory=None, filename=None, verbose=True):
    """

    Function to save data from a geopyv object. If no filename is
    provided, the host OS default file browser will be used to allow
    the user to choose a filename and storage location.


    Parameters
    ----------
    object : geopyv.object.Object
        The object to be saved.
    filename : str, optional
        The filename to give to the saved data file.


    .. note::
        * Any geopyv object can be passed to this function.
        * Do not include the .pyv extension in the `filename` argument.

    """
    cwd = os.getcwd()
    if directory is None and cwd not in filename:
        directory = os.getcwd()
    if filename is None:
        log.error(
            "No filename provided."
        )  # Add a method to select the filename here...
        return False
    if isinstance(object, gp.object.Object):
        solved = object.data["solved"]
        if solved is True or object.data["type"] == "Sequence":
            ext = ".pyv"
            filepath = directory + "/" + filename + ext
            with open(filepath, "wb") as outfile:
                if verbose:
                    message = "Saving geopyv object"
                    with alive_bar(dual_line=True, bar=None, title=message) as bar:
                        bar.text = "-> Saving object to {filepath}...".format(
                            filepath=filepath
                        )
                        pickle.dump(object.data, outfile)
                        bar()
                else:
                    pickle.dump(object.data, outfile)
                object_type = object.data["type"]
                if verbose:
                    log.info(
                        "Saved {object_type} object to {filepath}.".format(
                            object_type=object_type, filepath=filepath
                        )
                    )
                return True
        else:
            log.warn(
                "`{}` object not solved therefore no data to save.".format(
                    object.data["type"]
                )
            )
            return False
    else:
        log.error("Nothing saved. Object supplied not a valid geopyv type.")
        return False


def _img_loader(message, load):
    """

    Private method to open a file dialog and select an image.

    """
    directory = os.getcwd()
    dialog = gp.gui.selectors.image.ImageSelector()
    imgpath = dialog.get_path(directory, message)
    if load:
        return gp.image.Image(imgpath)
    else:
        return imgpath


def _load_f_img():
    """
    Private method to load the reference image.
    """
    log.warn("No reference image supplied. Please select the reference image.")
    return _img_loader("Select reference image.")


def _load_g_img():
    """
    Private method to load the target image.
    """
    log.warn("No target image supplied. Please select the target image.")
    return _img_loader("Select target image.")


def _load_img(name, load=True):
    """
    Private method to load an image.
    """
    log.warn("No {} image supplied. Please select the {} image.".format(name, name))
    return _img_loader("Select {} image.".format(name), load)


def _get_folder(message):
    """

    Private method to open a file dialog and select a folder.

    """
    directory = os.getcwd()
    dialog = gp.gui.selectors.folder.FolderSelector()
    folder_path = dialog.get_path(directory, message)
    return folder_path[len(directory) :]


def _get_image_dir():
    """

    Private method to get the `image_dir`.

    """
    log.warn("No `image_dir` supplied. Please select the `image_dir`.")
    return _get_folder("Select image directory.")


def _get_mesh_dir():
    """

    Private method to get the `mesh_dir`.

    """
    log.warn("No `mesh_folder` supplied. Please select the `mesh_folder`.")
    return _get_folder("Select mesh directory.")



"""

Subset module for geopyv.

"""
import logging
import cv2
import numpy as np
import geopyv as gp
from geopyv.object import Object
import traceback

log = logging.getLogger(__name__)


class SubsetBase(Object):
    """

    Subset base class to be used as a mixin. Contains plot functionality.

    """

    def __init__(self):
        super().__init__(object_type="Subset")
        """

        Subset base class initialiser.

        """

    def inspect(self, warp=False, show=True, block=True, save=None):
        """

        Method to show the subset and associated quality metrics using
        :mod:`~geopyv.plots.inspect_subset`.

        Parameters
        ----------
        show : bool, optional
            Control whether the plot is displayed.
        block : bool, optional
            Control whether the plot blocks execution until closed.
        save : str, optional
            Name to use to save plot. Uses default extension of `.png`.


        Returns
        -------
        fig :  `matplotlib.pyplot.figure`
            Figure object.
        ax : `matplotlib.pyplot.axes`
            Axes object.


        .. note::
            * The figure and axes objects can be returned allowing standard
              matplotlib functionality to be used to augment the plot generated.
              See the :ref:`plots tutorial <Plots Tutorial>` for guidance.

        .. seealso::
            :meth:`~geopyv.plots.inspect_subset`

        """

        # Check input.
        self._report(gp.check._check_type(warp, "warp", [bool]), "TypeError")
        if warp:
            # Check if solved.
            if self.data["solved"] is not True:
                log.error(
                    "Subset not yet solved therefore no warp data to plot. "
                    "First, run :meth:`~geopyv.subset.Subset.solve()` to solve."
                )
                raise ValueError(
                    "Mesh not yet solved therefore no warp data to plot. "
                    "First, run :meth:`~geopyv.subset.Subset.solve()` to solve."
                )
        self._report(gp.check._check_type(show, "show", [bool]), "TypeError")
        self._report(gp.check._check_type(block, "block", [bool]), "TypeError")
        self._report(gp.check._check_type(save, "save", [str, type(None)]), "TypeError")

        # Inspect subset.
        if warp:
            fig, ax = gp.plots.inspect_subset_warp(
                data=self.data,
                mask=None,
                show=show,
                block=block,
                save=save,
            )
        else:
            fig, ax = gp.plots.inspect_subset(
                data=self.data,
                mask=None,
                show=show,
                block=block,
                save=save,
            )
        return fig, ax

    def convergence(self, show=True, block=True, save=None):
        """

        Method to plot the rate of convergence for the subset using
        :mod:`~geopyv.plots.convergence_subset`.

        Parameters
        ----------
        show : bool, optional
            Control whether the plot is displayed.
        block : bool, optional
            Control whether the plot blocks execution until closed.
        save : str, optional
            Name to use to save plot. Uses default extension of `.png`.


        Returns
        -------
        fig :  matplotlib.pyplot.figure
            Figure object.
        ax : `matplotlib.pyplot.axes`
            Axes object.


        .. note::
            * The figure and axes objects can be returned allowing standard
              matplotlib functionality to be used to augment the plot generated.
              See the :ref:`plots tutorial <Plots Tutorial>` for guidance.

        .. warning::
            * Can only be used once the subset has been solved using the
              :meth:`~geopyv.subset.Subset.solve` method.

        .. seealso::
            :meth:`~geopyv.plots.convergence_subset`

        """

        # Check input.
        self._report(gp.check._check_type(show, "show", [bool]), "TypeError")
        self._report(gp.check._check_type(block, "block", [bool]), "TypeError")
        self._report(gp.check._check_type(save, "save", [str, type(None)]), "TypeError")

        # Plot convergence.
        if "results" in self.data:
            fig, ax = gp.plots.convergence_subset(
                data=self.data, show=show, block=block, save=save
            )
            return fig, ax
        else:
            log.error("Subset not yet solved. Run the solve() method.")

    def _report(self, msg, error_type):
        if msg and error_type != "Warning":
            log.error(msg)
        elif msg and error_type == "Warning":
            log.warning(msg)
            return True
        if error_type == "ValueError" and msg:
            raise ValueError(msg)
        elif error_type == "TypeError" and msg:
            raise TypeError(msg)
        elif error_type == "IndexError" and msg:
            raise IndexError(msg)


class Subset(SubsetBase):
    """

    Subset class for geopyv.

    Private Attributes
    ------------------
    _f_img : `geopyv.image.Image`
        Reference image of geopyv.image.Image class, instantiated by
        :mod:`~image.Image`.
    _g_img : `geopyv.image.Image`
        Target image of geopyv.image.Image class, instantiated by
        :mod:`~image.Image`.
    _template: `geopyv.templates.Template`
        Subset template object.
    _method : `str`
        Solver type. Options are 'ICGN' and 'FAGN'.
    _init_guess_size : int
        Size of subset used to define the initial guess, approximated
        by private method
        :meth:`~_get_initial_guess_size`.
    _f_coord : `numpy.ndarray` (x, y)
        1D array of the coordinates of the subset in reference image
        of type `float`.
    _f_coords : `numpy.ndarray` (Nx, 2)
        2D array of subset coordinates in reference image of type `float`.
    _grad_f : `numpy.ndarray` (Nx, 2)
        Gradients of reference image `f`.
    _SSSIG : float
        Sum of the square of the reference subset intensity gradients.
    _sigma_intensity : float
        Standard deviaition of the reference subset intensities.
    _p_0 : `numpy.ndarray` (Nx, 1)
        1D array of initial warp function parameters of type `float`,
        used to precondition class method :meth:`~solve`.
    _p : `numpy.ndarray` (Nx, 1)
        1D array of warp function parameters of type `float`, output by class
        method :meth:`~solve`.
    _norm : float
        Custom norm of the increment in the warp function parameters after
        Gao et al. (2015), computed by private method :meth:`~_get_norm`.
    _C_ZNSSD : float
        Zero-normalised sum of squared differences coefficient, computed by
        private method :meth:`~_get_correlation`.
    _C_ZNCC : float
        Zero-normalised cross-correlation coefficient, computed by private
        method :meth:`~_get_correlation`.
    _x : float
        Initial horizontal coordinate.
    _y : float
        Initial vertical coordinate.
    _u : float
        Horizontal displacement.
    _v : float
        Vertical displacement.
    _x_f : float
        Final horizontal coordinate.
    _y_f : float
        Final vertical coordinate.
    _settings: dict
        Dictionary of settings.
    _quality: dict
        Dictionary of image quality measures.
    _results: dict
        Dictionary of results.

    """

    def __init__(self, *, f_coord=None, f_img=None, g_img=None, template=None):
        """

        Initialisation of geopyv subset object.

        Parameters
        ----------
        coord : `numpy.ndarray` (x, y), optional
            Subset coordinates.
        f_img : geopyv.image.Image, optional
            Reference image of geopyv.image.Image class, instantiated by
            :mod:`~geopyv.image.Image`.
        g_img : geopyv.image.Image, optional
            Target image of geopyv.imageImage class, instantiated by
            :mod:`~geopyv.image.Image`.
        template : geopyv.templates.Template, optional
            Subset template object, instantiated by
            :mod:`~geopyv.templates.Circle` or :mod:`~geopyv.templates.Square`.


        Attributes
        ----------
        data : dict
            Data object containing all settings and results. See the data
            structure :ref:`here <subset_data_structure>`.
        solved : bool
            Boolean to indicate if the subset has been solved.

        """
        log.debug("Initialising geopyv Subset object.")

        # Check types.
        if self._report(
            gp.check._check_type(f_img, "f_img", [gp.image.Image]), "Warning"
        ):
            f_img = gp.io._load_f_img()
        if self._report(
            gp.check._check_type(g_img, "g_img", [gp.image.Image]), "Warning"
        ):
            g_img = gp.io._load_g_img()
        if template is None:
            template = gp.templates.Circle(50)
        types = [gp.templates.Circle, gp.templates.Square]
        self._report(gp.check._check_type(template, "template", types), "TypeError")
        if self._report(
            gp.check._check_type(f_coord, "f_coord", [np.ndarray]), "Warning"
        ):
            selector = gp.gui.selectors.coordinate.CoordinateSelector()
            f_coord = selector.select(f_img, template)
        elif self._report(gp.check._check_dim(f_coord, "f_coord", 1), "Warning"):
            selector = gp.gui.selectors.coordinate.CoordinateSelector()
            f_coord = selector.select(f_img, template)
        elif self._report(gp.check._check_axis(f_coord, "f_coord", 0, [2]), "Warning"):
            selector = gp.gui.selectors.coordinate.CoordinateSelector()
            f_coord = selector.select(f_img, template)

        # Store.
        self._initialised = False
        self._f_coord = f_coord
        self._f_img = f_img
        self._g_img = g_img
        self._template = template

        # Check subset is entirely within the reference image.
        self._x = self._f_coord[0]
        self._y = self._f_coord[1]
        subset_list = np.array(
            [
                [
                    self._x - self._template.size,
                    self._y - self._template.size,
                ],
                [
                    self._x + self._template.size,
                    self._y + self._template.size,
                ],
                [
                    self._x - self._template.size,
                    self._y + self._template.size,
                ],
                [
                    self._x + self._template.size,
                    self._y - self._template.size,
                ],
            ]
        )
        if np.any(subset_list < 0):
            raise ValueError(
                "Subset reference partially falls outside reference image."
            )

        # Initialise subset.
        self._get_initial_guess_size()
        output = gp._subset_extensions._init_reference(
            self._f_coord, self._template.coords, self._f_img.QCQT
        )
        self._f_coords = output[0]
        self._f = output[1]
        self._f_m = output[2][0][0]
        self._Delta_f = output[2][1][0]
        self._grad_f = output[3]
        self._SSSIG = output[4][0][0]
        self._sigma_intensity = output[4][1][0]
        self.solved = False
        self._unsolvable = False
        self._calibrated = False
        self._initialised = True
        self._quality = {
            "SSSIG": self._SSSIG,
            "sigma_intensity": self._sigma_intensity,
        }

        # Data.
        self.data = {
            "type": "Subset",
            "solved": self.solved,
            "unsolvable": self._unsolvable,
            "calibrated": self._calibrated,
            "images": {
                "f_img": self._f_img.filepath,
                "g_img": self._g_img.filepath,
            },
            "position": {
                "x": self._x,
                "y": self._y,
            },
            "quality": self._quality,
            "template": {
                "shape": self._template.shape,
                "dimension": self._template.dimension,
                "size": self._template.size,
                "n_px": self._template.n_px,
            },
        }
        log.debug("Initialised geopyv Subset object.")

    def solve(
        self,
        *,
        max_norm=1e-5,
        max_iterations=50,
        order=1,
        warp_0=np.zeros(6),
        tolerance=0.75,
        method="ICGN",
    ):
        r"""

        Method to solve for the subset displacements using the various methods.

        Parameters
        ----------
        max_norm : float, optional
            Exit criterion for norm of increment in warp function. Defaults to
            value of :math:`1 \cdot 10^{-3}`.
        max_iterations : int, optional
            Exit criterion for number of Gauss-Newton iterations. Defaults to
            value of 50.
        order : int
            Warp function order. Options are 1 and 2.
        warp_0 : ndarray, optional
            1D array of warp function parameters with `float` type.
        tolerance: float, optional
            Correlation coefficient tolerance. Defaults to a value of 0.7.
        method : str
            Solution method. Options are FAGN and ICGN. Default is ICGN since
            it is faster.


        Returns
        -------
        solved : `bool`
            Boolean to indicate if the subset instance has been solved.


        .. note::
            * The warp function parameter array can be used to precondition
              the computation if passed non-zero values.
            * Otherwise, the initial guess at the subset displacement is
              performed by :meth:`~_get_initial_guess`.
            * If not specified, the solver defaults to a first order warp
              function.
            * For guidance on how to use this class see the
              :ref:`subset tutorial <Subset Tutorial>`.


        .. seealso::
            :meth:`~_get_initial_guess_size`
            :meth:`~_get_initial_guess`

        """
        log.debug("Solving geopyv Subset object.")

        # Check input.
        check = gp.check._check_type(max_norm, "max_norm", [float])
        if check:
            try:
                max_norm = float(max_norm)
                self._report(
                    gp.check._conversion(max_norm, "max_norm", float), "Warning"
                )
            except Exception:
                self._report(check, "TypeError")
        self._report(gp.check._check_range(max_norm, "max_norm", 1e-20), "ValueError")
        check = gp.check._check_type(max_iterations, "max_iterations", [int])
        if check:
            try:
                max_iterations = int(max_iterations)
                self._report(
                    gp.check._conversion(max_iterations, "max_iterations", int),
                    "Warning",
                )
            except Exception:
                self._report(check, "TypeError")
        self._report(
            gp.check._check_range(max_iterations, "max_iterations", 0.5), "ValueError"
        )
        check = gp.check._check_type(tolerance, "tolerance", [float])
        if check:
            try:
                tolerance = float(tolerance)
                self._report(
                    gp.check._conversion(tolerance, "tolerance", float), "Warning"
                )
            except Exception:
                self._report(check, "TypeError")
        self._report(
            gp.check._check_range(tolerance, "tolerance", 0.0, 1.0), "ValueError"
        )
        check = gp.check._check_type(order, "order", [int])
        if check:
            try:
                order = int(order)
                self._report(gp.check._conversion(order, "order", int), "Warning")
            except Exception:
                self._report(check, "TypeError")
        self._report(gp.check._check_value(order, "order", [1, 2]), "ValueError")
        check = gp.check._check_type(warp_0, "warp_0", [np.ndarray])
        if check:
            try:
                warp_0 = np.asarray(warp_0)
                self._report(
                    gp.check._conversion(warp_0, "warp_0", np.ndarray), "Warning"
                )
            except Exception:
                self._report(check, "TypeError")
        self._report(gp.check._check_dim(warp_0, "warp_0", 1), "ValueError")
        self._report(
            gp.check._check_axis(warp_0, "warp_0", 0, [6 * order]), "ValueError"
        )

        # Store settings.
        self._method = method
        self._order = order
        self._max_norm = max_norm
        self._max_iterations = max_iterations
        self._tolerance = tolerance
        self._settings = {
            "method": self._method,
            "order": self._order,
            "max_norm": self._max_norm,
            "max_iterations": self._max_iterations,
            "tolerance": self._tolerance,
        }
        self.data.update({"settings": self._settings})

        # Compute initial guess if u and v in deformation parameter vector
        # initialised with zeros, otherwise precondition.
        if warp_0[0] == 0 and warp_0[1] == 0:
            self._p_init = warp_0
            self._get_initial_guess()
            self._p = self._p_init
        else:
            self._p = warp_0
        # Call appropriate iterative solver:
        try:
            if self._method == "ICGN" and np.mod(self._p.shape[0], 2) == 0:
                results = gp._subset_extensions._solve_ICGN(
                    self._f_coord,
                    self._f_coords,
                    self._f,
                    self._f_m,
                    self._Delta_f,
                    self._grad_f,
                    self._f_img.QCQT,
                    self._g_img.QCQT,
                    self._p,
                    self._max_norm,
                    self._max_iterations,
                )
            elif self._method == "FAGN" and np.mod(self._p.shape[0], 2) == 0:
                results = gp._subset_extensions._solve_FAGN(
                    self._f_coord,
                    self._f_coords,
                    self._f,
                    self._f_m,
                    self._Delta_f,
                    self._grad_f,
                    self._f_img.QCQT,
                    self._g_img.QCQT,
                    self._p,
                    self._max_norm,
                    self._max_iterations,
                )
            # Unpack results.
            self._g_coords = results[0]
            self._g = results[1]
            self._g_m = results[2][0][0]
            self._g_m = results[2][1][0]
            self._iterations = np.max(results[3][0, :]).astype(int)
            self._history = results[3][:, : self._iterations]
            self._norm = self._history[1][-1]
            self._C_ZNCC = self._history[2][-1]
            self._C_ZNSSD = self._history[3][-1]
            self._p = results[4]
            self._u = self._p[0][0]
            self._v = self._p[1][0]
            self._x_f = self._x + self._u
            self._y_f = self._y + self._v

            # Check for tolerance.
            if self._C_ZNCC > self._tolerance:
                self.solved = True
            else:
                self.solved = False
            # Pack results.
            self._results = {
                "u": self._u,
                "v": self._v,
                "p": self._p,
                "history": self._history,
                "iterations": self._iterations,
                "norm": self._norm,
                "C_ZNCC": self._C_ZNCC,
                "C_ZNSSD": self._C_ZNSSD,
            }
            self.data.update({"results": self._results})
        except Exception:
            log.error("Subset unsolvable.")
            print(traceback.format_exc())
            self.solved = False
            self._unsolvable = True

        # Pack results.
        self.data.update(
            {
                "solved": self.solved,
                "unsolvable": self._unsolvable,
            }
        )

        return self.solved

    def _get_initial_guess_size(self):
        """

        Private method to estimate the size of square subset to use in the
        initial guess.

        """
        self._initial_guess_size = np.round(np.sqrt(self._template.n_px), 1)

    def _get_initial_guess(self):
        """

        Private method to compute an initial guess of the subset displacement
        using OpenCV function :py:meth:`cv2.matchTemplate` and the Normalised
        Cross-Correlation (NCC) criteria.

        """
        # Extract square subset for initial guess.
        x = self._f_coord.item(0)
        y = self._f_coord.item(1)
        x_min = (np.round(x, 0) - self._initial_guess_size / 2).astype(int)
        x_max = (np.round(x, 0) + self._initial_guess_size / 2).astype(int)
        y_min = (np.round(y, 0) - self._initial_guess_size / 2).astype(int)
        y_max = (np.round(y, 0) + self._initial_guess_size / 2).astype(int)
        subset = self._f_img.image_gs.astype(np.float32)[y_min:y_max, x_min:x_max]

        # Apply template matching technique.
        res = cv2.matchTemplate(
            self._g_img.image_gs.astype(np.float32),
            subset,
            cv2.TM_CCORR_NORMED,
        )
        max_loc = cv2.minMaxLoc(res)[3]

        # Initialised warp vector with preconditioned affine displacements.
        self._p_init[0] = (max_loc[0] + self._initial_guess_size / 2) - x
        self._p_init[1] = (max_loc[1] + self._initial_guess_size / 2) - y


class SubsetResults(SubsetBase):
    def __init__(self, data):
        """

        Subset results object for geopyv.

        Parameters
        ----------
        data : dict
            geopyv data dict from Subset object.


        Attributes
        ----------
        data : dict
            geopyv data dict from Subset object.


        .. note::
            * Contains all of the plot functionality provied by
              :class:`~geopyv.subset.SubsetBase` but none of the algorithms
              provided by :class:`~geopyv.subset.Subset` (i.e. you can't use
              this to re-analyse images). Purely used to store data and
              interrogate results.

        .. warning::
            * To re-analyse data instantiate a new object using
              :class:`~geopyv.subset.Subset` and use the
              :class:`~geopyv.subset.Subset.solve` method.

        """
        self.data = data



"""

Mesh module for geopyv.

"""
import logging
import numpy as np
import geopyv as gp
import scipy as sp
from geopyv.object import Object
import gmsh
from copy import deepcopy
from scipy.optimize import minimize_scalar
from alive_progress import alive_bar
import traceback

log = logging.getLogger(__name__)


class MeshBase(Object):
    """

    Mesh base class to be used as a mixin. Contains plot functionality.

    """

    def __init__(self):
        super().__init__(object_type="Mesh")
        """

        Mesh base class initialiser.

        """

    def inspect(self, *, subset_index=None, show=True, block=True, save=None):
        """

        Method to show the mesh and associated subset quality metrics using
        :mod: `~geopyv.plots.inspect_subset` for subsets and
        :mod: `~geopyv.plots.inspect_mesh` for meshes.

        Parameters
        ----------
        subset_index : int, optional
            Index of the subset to inspect. If `None', the mesh is inspected instead.
        show : bool, optional
            Control whether the plot is displayed.
        block : bool, optional
            Control whether the plot blocks execution until closed.
        save : str, optional
            Name to use to save plot. Uses default extension of `.png`.

        Returns
        -------
        fig :  `matplotlib.pyplot.figure`
            Figure object.
        ax : `matplotlib.pyplot.axes`
            Axes object.

        .. note::
            * The figure and axes objects can be returned allowing standard
              matplotlib functionality to be used to augment the plot generated.
              See the :ref:`plots tutorial <Plots Tutorial>` for guidance.

        .. seealso::
            :meth:`~geopyv.plots.inspect_subset`
            :meth:`~geopyv.plots.inspect_mesh`

        """

        # Check input.
        self._report(
            gp.check._check_type(subset_index, "subset_index", [int, type(None)]),
            "TypeError",
        )
        if subset_index:
            self._report(
                gp.check._check_index(
                    subset_index, "subset_index", 0, self.data["nodes"]
                ),
                "IndexError",
            )
        self._report(gp.check._check_type(show, "show", [bool]), "TypeError")
        self._report(gp.check._check_type(block, "block", [bool]), "TypeError")
        self._report(gp.check._check_type(save, "save", [str, type(None)]), "TypeError")

        # Inspect.
        if subset_index is not None:
            if self.data["solved"]:
                subset_data = self.data["results"]["subsets"][subset_index]
            else:
                subset_data = gp.subset.Subset(
                    f_coord=self._nodes[subset_index],
                    f_img=self._f_img,
                    g_img=self._g_img,
                    template=self._template,
                ).data
            mask = self.data["mask"]
            log.info("Inspecting subset {subset}...".format(subset=subset_index))
            fig, ax = gp.plots.inspect_subset(
                data=subset_data,
                mask=mask,
                show=show,
                block=block,
                save=save,
            )
            return fig, ax
        else:
            log.info("Inspecting mesh...")
            fig, ax = gp.plots.inspect_mesh(
                data=self.data,
                show=show,
                block=block,
                save=save,
            )
            return fig, ax

    def convergence(
        self,
        *,
        subset_index=None,
        quantity=None,
        show=True,
        block=True,
        save=None,
    ):
        """

        Method to plot the rate of convergence using
        :mod: `~geopyv.plots.convergence_subset` for subsets and
        :mod: `~geopyv.plots.convergence_mesh` for meshes.

        Parameters
        ----------
        subset_index : int, optional
            Index of the subset to inspect.
            If `None', the convergence plot is for the mesh instead.
        quantity : str, optional
            Selector for histogram convergence property if the convergence
            plot is for mesh. Defaults to `C_ZNCC` if left as default None.
        show : bool, optional
            Control whether the plot is displayed.
        block : bool, optional
            Control whether the plot blocks execution until closed.
        save : str, optional
            Name to use to save plot. Uses default extension of `.png`.


        Returns
        -------
        fig :  matplotlib.pyplot.figure
            Figure object.
        ax : `matplotlib.pyplot.axes`
            Axes object.


        .. note::
            * The figure and axes objects can be returned allowing standard
              matplotlib functionality to be used to augment the plot generated.
              See the :ref:`plots tutorial <Plots Tutorial>` for guidance.

        .. warning::
            * Can only be used once the mesh has been solved using the
              :meth:`~geopyv.mesh.Mesh.solve` method.

        .. seealso::
            :mod: `~geopyv.plots.convergence_subset`
            :meth:`~geopyv.plots.convergence_mesh`

        """

        # Check if solved.
        if self.data["solved"] is not True:
            log.error(
                "Mesh not yet solved therefore no convergence data to plot. "
                "First, run :meth:`~geopyv.mesh.Mesh.solve()` to solve."
            )
            raise ValueError(
                "Mesh not yet solved therefore no convergence data to plot. "
                "First, run :meth:`~geopyv.mesh.Mesh.solve()` to solve."
            )

        # Check input.
        self._report(
            gp.check._check_type(subset_index, "subset_index", [int, type(None)]),
            "TypeError",
        )
        if subset_index:
            self._report(
                gp.check._check_index(
                    subset_index, "subset_index", 0, self.data["nodes"]
                ),
                "IndexError",
            )
        self._report(
            gp.check._check_type(quantity, "quantity", [str, type(None)]), "TypeError"
        )
        if quantity:
            self._report(
                gp.check._check_value(
                    quantity, "quantity", ["C_ZNCC", "iterations", "norm"]
                ),
                "ValueError",
            )
        self._report(gp.check._check_type(show, "show", [bool]), "TypeError")
        self._report(gp.check._check_type(block, "block", [bool]), "TypeError")
        self._report(gp.check._check_type(save, "save", [str, type(None)]), "TypeError")

        # Plot convergence.
        if subset_index is not None:
            log.info(
                "Generating convergence plots for subset {subset}...".format(
                    subset=subset_index
                )
            )
            fig, ax = gp.plots.convergence_subset(
                self.data["results"]["subsets"][subset_index],
                show=show,
                block=block,
                save=save,
            )
            return fig, ax
        else:
            if quantity is None:
                quantity = "C_ZNCC"
            log.info(
                "Generating {quantity} convergence histogram for mesh...".format(
                    quantity=quantity
                )
            )
            fig, ax = gp.plots.convergence_mesh(
                data=self.data,
                quantity=quantity,
                show=show,
                block=block,
                save=save,
            )
            return fig, ax

    def contour(
        self,
        *,
        quantity="C_ZNCC",
        imshow=True,
        colorbar=True,
        ticks=None,
        mesh=False,
        alpha=0.75,
        levels=None,
        axis=True,
        xlim=None,
        ylim=None,
        show=True,
        block=True,
        save=None,
    ):
        """

        Method to plot the contours of a given quantity.

        Parameters
        ----------
        quantity : str, optional
            Selector for contour parameter. Must be in:
            [`C_ZNCC`, `iterations`, `norm`, `u`, `v`, `u_x`, `v_x`,`u_y`,`v_y`, `R`]
            Defaults to `C_ZNCC`.
        imshow : bool, optional
            Control whether the reference image is plotted.
            Defaults to True.
        colorbar : bool, optional
            Control whether the colour bar is plotted.
            Defaults to True.
        ticks : list, optional
            Overwrite default colourbar ticks.
            Defaults to None.
        mesh : bool, optional
            Control whether the mesh is plotted.
            Defaults to False.
        alpha : float, optional
            Control contour opacity. Must be between 0.0-1.0.
            Defaults to 0.75.
        levels : int or array-like, optional
            Control number and position of contour lines.
            Defaults to None.
        axis : bool, optional
            Control whether the axes are plotted.
            Defaults to True.
        xlim : array-like, optional
            Set the plot x-limits (lower_limit,upper_limit).
            Defaults to None.
        ylim : array-like, optional
            Set the plot y-limits (lower_limit,upper_limit).
            Defaults to None.
        show : bool, optional
            Control whether the plot is displayed.
            Defaults to True.
        block : bool, optional
            Control whether the plot blocks execution until closed.
            Defaults to False.
        save : str, optional
            Name to use to save plot. Uses default extension of `.png`.


        Returns
        -------
        fig :  matplotlib.pyplot.figure
            Figure object.
        ax : `matplotlib.pyplot.axes`
            Axes object.


        .. note::
            * The figure and axes objects can be returned allowing standard
              matplotlib functionality to be used to augment the plot generated.
              See the :ref:`plots tutorial <Plots Tutorial>` for guidance.

        .. warning::
            * Can only be used once the mesh has been solved using the
              :meth:`~geopyv.mesh.Mesh.solve` method.

        .. seealso::
            :meth:`~geopyv.plots.contour_mesh`

        """
        # Check if solved.
        if self.data["solved"] is not True:
            log.error(
                "Mesh not yet solved therefore no convergence data to plot. "
                "First, run :meth:`~geopyv.mesh.Mesh.solve()` to solve."
            )
            raise ValueError(
                "Mesh not yet solved therefore no convergence data to plot. "
                "First, run :meth:`~geopyv.mesh.Mesh.solve()` to solve."
            )

        # Check input.
        self._report(
            gp.check._check_type(quantity, "quantity", [str, type(None)]), "TypeError"
        )
        types = [
            "C_ZNCC",
            "iterations",
            "norm",
            "u",
            "v",
            "u_x",
            "v_x",
            "u_y",
            "v_y",
            "R",
        ]
        if quantity:
            self._report(
                gp.check._check_value(quantity, "quantity", types), "ValueError"
            )
        self._report(gp.check._check_type(imshow, "imshow", [bool]), "TypeError")
        self._report(gp.check._check_type(colorbar, "colorbar", [bool]), "TypeError")
        types = [tuple, list, np.ndarray, type(None)]
        self._report(gp.check._check_type(ticks, "ticks", types), "TypeError")
        self._report(gp.check._check_type(mesh, "mesh", [bool]), "TypeError")
        check = gp.check._check_type(alpha, "alpha", [float])
        if check:
            try:
                alpha = float(alpha)
                self._report(gp.check._conversion(alpha, "alpha", float), "Warning")
            except Exception:
                self._report(check, "TypeError")
        self._report(gp.check._check_range(alpha, "alpha", 0.0, 1.0), "ValueError")
        types = [int, tuple, list, np.ndarray, type(None)]
        self._report(gp.check._check_type(levels, "levels", types), "TypeError")
        self._report(gp.check._check_type(axis, "axis", [bool]), "TypeError")
        types = [tuple, list, np.ndarray, type(None)]
        self._report(gp.check._check_type(xlim, "xlim", types), "TypeError")
        if xlim is not None:
            self._report(gp.check._check_dim(xlim, "xlim", 1), "ValueError")
            self._report(gp.check._check_axis(xlim, "xlim", 0, [2]), "ValueError")
        self._report(gp.check._check_type(ylim, "ylim", types), "TypeError")
        if ylim is not None:
            self._report(gp.check._check_dim(ylim, "ylim", 1), "ValueError")
            self._report(gp.check._check_axis(ylim, "ylim", 0, [2]), "ValueError")
        self._report(gp.check._check_type(show, "show", [bool]), "TypeError")
        self._report(gp.check._check_type(block, "block", [bool]), "TypeError")
        self._report(gp.check._check_type(save, "save", [str, type(None)]), "TypeError")

        # Plot contours.
        log.info(
            "Generating {quantity} contour plot for mesh...".format(quantity=quantity)
        )
        fig, ax = gp.plots.contour_mesh(
            data=self.data,
            imshow=imshow,
            quantity=quantity,
            colorbar=colorbar,
            ticks=ticks,
            mesh=mesh,
            alpha=alpha,
            levels=levels,
            axis=axis,
            xlim=xlim,
            ylim=ylim,
            show=show,
            block=block,
            save=save,
        )
        return fig, ax

    def quiver(
        self,
        *,
        scale=1.0,
        imshow=True,
        mesh=False,
        axis=True,
        xlim=None,
        ylim=None,
        show=True,
        block=True,
        save=None,
    ):
        """

        Method to plot a quiver plot of the displacements.

        Parameters
        ----------
        scale : float, optional
            Control size of quiver arrows.
        imshow : bool, optional
            Control whether the reference image is plotted.
            Defaults to True.
        mesh : bool, optional
            Control whether the mesh is plotted.
            Defaults to False.
        axis : bool, optional
            Control whether the axes are plotted.
            Defaults to True.
        xlim : array-like, optional
            Set the plot x-limits (lower_limit,upper_limit).
            Defaults to None.
        ylim : array-like, optional
            Set the plot y-limits (lower_limit,upper_limit).
            Defaults to None.
        show : bool, optional
            Control whether the plot is displayed.
            Defaults to True.
        block : bool, optional
            Control whether the plot blocks execution until closed.
            Defaults to False.
        save : str, optional
            Name to use to save plot. Uses default extension of `.png`.

        Returns
        -------
        fig :  matplotlib.pyplot.figure
            Figure object.
        ax : `matplotlib.pyplot.axes`
            Axes object.


        .. note::
            * The figure and axes objects can be returned allowing standard
              matplotlib functionality to be used to augment the plot generated.
              See the :ref:`plots tutorial <Plots Tutorial>` for guidance.

        .. warning::
            * Can only be used once the mesh has been solved using the
              :meth:`~geopyv.mesh.Mesh.solve` method.

        .. seealso::
            :meth:`~geopyv.plots.quiver_mesh`
        """

        # Check if solved.
        if self.data["solved"] is not True:
            log.error(
                "Mesh not yet solved therefore no convergence data to plot. "
                "First, run :meth:`~geopyv.mesh.Mesh.solve()` to solve."
            )
            raise ValueError(
                "Mesh not yet solved therefore no convergence data to plot. "
                "First, run :meth:`~geopyv.mesh.Mesh.solve()` to solve."
            )

        # Check inputs.
        check = gp.check._check_type(scale, "scale", [float])
        if check:
            try:
                scale = float(scale)
                self._report(gp.check._conversion(scale, "scale", float), "Warning")
            except Exception:
                self._report(check, "TypeError")
        self._report(gp.check._check_range(scale, "scale", 0.0), "ValueError")
        self._report(gp.check._check_type(imshow, "imshow", [bool]), "TypeError")
        self._report(gp.check._check_type(mesh, "mesh", [bool]), "TypeError")
        self._report(gp.check._check_type(axis, "axis", [bool]), "TypeError")
        types = [tuple, list, np.ndarray, type(None)]
        self._report(gp.check._check_type(xlim, "xlim", types), "TypeError")
        if xlim is not None:
            self._report(gp.check._check_dim(xlim, "xlim", 1), "ValueError")
            self._report(gp.check._check_axis(xlim, "xlim", 0, [2]), "ValueError")
        self._report(gp.check._check_type(ylim, "ylim", types), "TypeError")
        if ylim is not None:
            self._report(gp.check._check_dim(ylim, "ylim", 1), "ValueError")
            self._report(gp.check._check_axis(ylim, "ylim", 0, [2]), "ValueError")
        self._report(gp.check._check_type(show, "show", [bool]), "TypeError")
        self._report(gp.check._check_type(block, "block", [bool]), "TypeError")
        self._report(gp.check._check_type(save, "save", [str, type(None)]), "TypeError")

        # Plot quiver.
        log.info("Generating quiver plot for mesh...")
        fig, ax = gp.plots.quiver_mesh(
            data=self.data,
            scale=scale,
            imshow=imshow,
            mesh=mesh,
            axis=axis,
            xlim=xlim,
            ylim=ylim,
            show=show,
            block=block,
            save=save,
        )
        return fig, ax

    def _report(self, msg, error_type):
        if msg and error_type != "Warning":
            log.error(msg)
        elif msg and error_type == "Warning":
            log.warning(msg)
            return True
        if error_type == "ValueError" and msg:
            raise ValueError(msg)
        elif error_type == "TypeError" and msg:
            raise TypeError(msg)
        elif error_type == "IndexError" and msg:
            raise IndexError(msg)


class Mesh(MeshBase):
    """
    Mesh class for geopyv.

    Private Attributes
    ------------------
    _f_img : `geopyv.image.Image`
        Reference image of geopyv.image.Image class, instantiated by
        :mod:`~image.Image`.
    _g_img : `geopyv.image.Image`
        Target image of geopyv.image.Image class, instantiated by
        :mod:`~image.Image`.
    _target_nodes : int
        Target number of nodes. Defaults to a value of 1000.
    _boundary_obj : `geopyv.geometry.region.Circle` or `geopyv.geometry.region.Path`
        geopyv.geometry.region.Region class defining the boundary geometry and
        behaviours.
    _exclusion_objs : list
        List of `geopyv.geometry.region.Region` class objects defining the
        exclusion geometry and behaviours.
    _size_lower_bound : float
        Lower bound on element size.
    _size_upper_bound : float
        Lower bound on element size.
    _mesh_order : int
        Mesh element order. Either 1 or 2.
    _unsolvable : bool
        Boolean to indicate if the mesh cannot be solved.
    _nodes : `numpy.ndarray` (Nx,2)
        Array of subset coordinates.
    _elements : `numpy.ndarray(Nx,3*_mesh_order)
        Node connectivity matrix.

    """

    def __init__(
        self,
        *,
        f_img=None,
        g_img=None,
        target_nodes=1000,
        boundary_obj=None,
        exclusion_objs=[],
        size_lower_bound=1.0,
        size_upper_bound=1000.0,
        mesh_order=2,
    ):
        """

        Initialisation of geopyv mesh object.

        Parameters
        ----------
        f_img : geopyv.image.Image, optional
            Reference image of geopyv.image.Image class, instantiated
            by :mod:`~geopyv.image.Image`.
        g_img : geopyv.image.Image, optional
            Target image of geopyv.imageImage class, instantiated by
            :mod:`~geopyv.image.Image`.
        target_nodes : int, optional
            Target number of nodes. Defaults to a value of 1000.
        boundary_obj : geopyv.geometry.region.Region
            Boundary of geopyv.geometry.region.Region class, instantiated
            by :mod: `~geopyv.geometry.region.Region` to define the boundary
            of the region of interest (RoI).
        exclusion_objs : list, optional
            List of geopyv.geometry.region.Region class objects, instantiated
            by :mod: `~geopyv.geometry.region.Region` to define exclusion regions
            within the region of interest (RoI).
        size_lower_bound : float, optional
            Lower bound on element size. Defaults to a value of 1.0.
        size_upper_bound : float, optional
            Lower bound on element size. Defaults to a value of 1000.0.
        mesh_order : int, optional
            Mesh element order. Options are 1 and 2. Defaults to 2.

        Attributes
        ----------
        data : dict
            Data object containing all settings and results.
            See the data structure :ref:`here <mesh_data_structure>`.
        solved : bool
            Boolean to indicate if the mesh has been solved.

        """

        # Set initialised boolean.
        self.initialised = False

        # Check types.
        if self._report(
            gp.check._check_type(f_img, "f_img", [gp.image.Image]), "Warning"
        ):
            f_img = gp.io._load_f_img()
        if self._report(
            gp.check._check_type(g_img, "g_img", [gp.image.Image]), "Warning"
        ):
            g_img = gp.io._load_g_img()
        check = gp.check._check_type(target_nodes, "target_nodes", [int])
        if check:
            try:
                target_nodes = int(target_nodes)
                self._report(
                    gp.check._conversion(target_nodes, "target_nodes", int), "Warning"
                )
            except Exception:
                self._report(check, "TypeError")
        self._report(
            gp.check._check_range(target_nodes, "target_nodes", 1), "ValueError"
        )
        self._report(
            gp.check._check_type(
                boundary_obj,
                "boundary_obj",
                [gp.geometry.region.Circle, gp.geometry.region.Path],
            ),
            "TypeError",
        )
        check = gp.check._check_type(exclusion_objs, "exclusion_objs", [list])
        if check:
            try:
                exclusion_objs = list(exclusion_objs)
                self._report(
                    gp.check._conversion(exclusion_objs, "exclusion_objs", list, False),
                    "Warning",
                )
            except Exception:
                self._report(check, "TypeError")
        for exclusion_obj in exclusion_objs:
            self._report(
                gp.check._check_type(
                    exclusion_obj,
                    "exclusion_obj",
                    [gp.geometry.region.Circle, gp.geometry.region.Path],
                ),
                "TypeError",
            )
        self._report(
            gp.check._check_type(size_lower_bound, "size_lower_bound", [int, float]),
            "TypeError",
        )
        self._report(
            gp.check._check_type(size_upper_bound, "size_upper_bound", [int, float]),
            "TypeError",
        )
        self._report(
            gp.check._check_comp(
                size_lower_bound,
                "size_lower_bound",
                size_upper_bound,
                "size_upper_bound",
            ),
            "ValueError",
        )
        check = gp.check._check_type(mesh_order, "mesh_order", [int])
        if check:
            try:
                mesh_order = int(mesh_order)
                self._report(
                    gp.check._conversion(mesh_order, "mesh_order", int), "Warning"
                )
            except Exception:
                self._report(check, "TypeError")
        self._report(
            gp.check._check_value(mesh_order, "mesh_order", [1, 2]), "ValueError"
        )

        # Store.
        self._initialised = False
        self._f_img = f_img
        self._g_img = g_img
        self._target_nodes = target_nodes
        self._boundary_obj = boundary_obj
        if self._boundary_obj._option == "R":
            self._boundary_obj._option = "F"
        self._exclusion_objs = exclusion_objs
        self._size_lower_bound = size_lower_bound
        self._size_upper_bound = size_upper_bound
        self._mesh_order = mesh_order
        self._hard_boundary = boundary_obj._hard
        self.solved = False
        self._unsolvable = False
        self._calibrated = False

        # Checks.
        self._target_checks()  # Size bounds and target nodes.
        self._update_region()  # Boundary and exclusion objects.

        # Creating the initial mesh.
        # Define region of interest.
        (
            self._borders,
            self._segments,
            self._curves,
            self._mask,
        ) = gp.geometry.meshing._define_RoI(
            self._f_img, self._boundary_obj, self._exclusion_objs
        )
        # Initialize gmsh if not already initialized.
        if gmsh.isInitialized() == 0:
            gmsh.initialize()
        gmsh.option.setNumber("General.Verbosity", 2)
        # 0: silent except for fatal errors, 1: +errors, 2: +warnings,
        # 3: +direct, 4: +information, 5: +status, 99: +debug.
        log.info(
            "Generating mesh using gmsh with approximately {n} nodes.".format(
                n=self._target_nodes
            )
        )
        self._initial_mesh()
        self._update_mesh()
        log.info(
            "Mesh generated with {n} nodes and {e} elements.".format(
                n=len(self._nodes), e=len(self._elements)
            )
        )
        gmsh.finalize()

        # Data.
        self.data = {
            "type": "Mesh",
            "solved": self.solved,
            "unsolvable": self._unsolvable,
            "calibrated": self._calibrated,
            "images": {
                "f_img": self._f_img.filepath,
                "g_img": self._g_img.filepath,
            },
            "target_nodes": self._target_nodes,
            "mesh_order": self._mesh_order,
            "boundary_obj": self._boundary_obj,
            "exclusion_objs": self._exclusion_objs,
            "boundary": self._boundary,
            "exclusions": self._exclusions,
            "size_lower_bound": self._size_lower_bound,
            "size_upper_bound": self._size_upper_bound,
            "nodes": self._nodes,
            "elements": self._elements,
            "mask": self._mask,
        }

        self.initialised = True

    def _target_checks(self):
        """Private method to limit the mesh size upper bound by the boundary geometry
        and the target nodes by the number included in the boundary and exclusion
        geometries."""

        self._size_upper_bound = min(
            self._size_upper_bound,
            np.max(
                np.sqrt(
                    np.sum(
                        np.square(np.diff(self._boundary_obj._nodes, axis=0)), axis=1
                    )
                )
            ),
        )
        minimum_nodes = np.shape(self._boundary_obj._nodes)[0]
        for exclusion_obj in self._exclusion_objs:
            minimum_nodes += np.shape(exclusion_obj._nodes)[0]
        self._target_nodes = max(self._target_nodes, minimum_nodes)

    def set_target_nodes(self, target_nodes):
        """

        Method to create a mesh with a target number of nodes.

        Parameters
        ----------
        target_nodes : int
            Target number of nodes.


        .. note::
            * This method can be used to update the number of target nodes.
            * It will generate a new initial mesh with the specified target
              number of nodes.

        """

        # Check inputs.
        check = gp.check._check_type(target_nodes, "target_nodes", [int])
        if check:
            try:
                target_nodes = int(target_nodes)
                self._report(
                    gp.check._conversion(target_nodes, "target_nodes", int), "Warning"
                )
            except Exception:
                self._report(check, "TypeError")

        # Store.
        self._target_nodes = target_nodes

        # Checks.
        self._target_checks()  # Size bounds and target nodes.

        # Initialize gmsh if not already initialized.
        if gmsh.isInitialized() == 0:
            gmsh.initialize()
        gmsh.option.setNumber("General.Verbosity", 2)

        # Create mesh.
        log.info(
            "Generating mesh using gmsh with approximately {n} nodes.".format(
                n=self._target_nodes
            )
        )
        self._initial_mesh()
        self._update_mesh()
        log.info(
            "Mesh generated with {n} nodes and {e} elements.".format(
                n=len(self._nodes), e=len(self._elements)
            )
        )
        gmsh.finalize()

    def solve(
        self,
        *,
        template=None,
        seed_coord=None,
        seed_warp=None,
        max_norm=1e-5,
        max_iterations=50,
        subset_order=1,
        tolerance=0.75,
        seed_tolerance=0.9,
        method="ICGN",
        adaptive_iterations=0,
        correction=True,
        alpha=0.5,
        history=None,
        override=False,
    ):
        r"""

        Method to solve for the mesh.

        Parameters
        ----------
        template : `geopyv.templates.Template`
            Subset template object.
        seed_coord : `numpy.ndarray` (2,)
            An image coordinate selected in a region of low deformation. The
            reliability-guided approach is initiated from the nearest subset.
        seed_warp : `numpy.ndarray` (6*subset_order,)
            Deformation preconditioning vector for the seed subset.
        max_norm : float, optional
            Exit criterion for norm of increment in warp function.
            Defaults to value of
            :math:`1 \cdot 10^{-5}`.
        max_iterations : int, optional
            Exit criterion for number of Gauss-Newton iterations.
            Defaults to value
            of 50.
        subset_order : int
            Warp function order. Options are 1 and 2.
            Defaults to a value of 1.
        tolerance: float, optional
            Correlation coefficient tolerance.
            Defaults to a value of 0.75.
        seed_tolerance: float, optional
            Correlation coefficient tolerance for the seed subset.
            Defaults to a value of 0.9.
        method : str
            Solution method. Options are FAGN and ICGN.
            Default is ICGN since it is faster.
        adaptive_iterations : int, optional
            Number of mesh adaptivity iterations to perform.
            Defaults to a value of 0.
        correction : bool, optional
            Boolean indicator as to whether to correct poor
            correlation subsets.
            Defaults to a value of True.
        alpha : float, optional
            Mesh adaptivity control parameter.
            Defaults to a value of 0.5.

        Returns
        -------
        solved : bool
            Boolean to indicate if the mesh instance has been solved.

        .. note::
            * For guidance on how to use this class see the
              :ref:`mesh tutorial <Mesh Tutorial>`.

        """

        # Check if solved.
        if self.data["solved"] is True:
            log.error("Mesh has already been solved. Cannot be solved again.")
            return

        # Check inputs.
        if template is None:
            template = gp.templates.Circle(50)
        types = [gp.templates.Circle, gp.templates.Square]
        self._report(gp.check._check_type(template, "template", types), "TypeError")
        if self._report(
            gp.check._check_type(seed_coord, "seed_coord", [np.ndarray]), "Warning"
        ):
            selector = gp.gui.selectors.coordinate.CoordinateSelector()
            seed_coord = selector.select(self._f_img, template)
        elif self._report(gp.check._check_dim(seed_coord, "seed_coord", 1), "Warning"):
            selector = gp.gui.selectors.coordinate.CoordinateSelector()
            seed_coord = selector.select(self._f_img, template)
        elif self._report(
            gp.check._check_axis(seed_coord, "seed_coord", 0, [2]), "Warning"
        ):
            selector = gp.gui.selectors.coordinate.CoordinateSelector()
            seed_coord = selector.select(self._f_img, template)
        if seed_warp is not None:
            check = gp.check._check_type(seed_warp, "seed_warp", [np.ndarray])
            if check:
                try:
                    seed_warp = np.asarray(seed_warp)
                    self._report(
                        gp.check._conversion(seed_warp, "seed_warp", np.ndarray),
                        "Warning",
                    )
                except Exception:
                    self._report(check, "TypeError")
            self._report(gp.check._check_dim(seed_warp, "seed_warp", 1), "ValueError")
            self._report(
                gp.check._check_axis(seed_warp, "seed_warp", 0, [6 * subset_order]),
                "ValueError",
            )
        else:
            seed_warp = np.zeros(6 * subset_order)
        check = gp.check._check_type(max_norm, "max_norm", [float])
        if check:
            try:
                max_norm = float(max_norm)
                self._report(
                    gp.check._conversion(max_norm, "max_norm", float), "Warning"
                )
            except Exception:
                self._report(check, "TypeError")
        self._report(gp.check._check_range(max_norm, "max_norm", 1e-20), "ValueError")
        check = gp.check._check_type(max_iterations, "max_iterations", [int])
        if check:
            try:
                max_iterations = int(max_iterations)
                self._report(
                    gp.check._conversion(max_iterations, "max_iterations", int),
                    "Warning",
                )
            except Exception:
                self._report(check, "TypeError")
        self._report(
            gp.check._check_range(max_iterations, "max_iterations", 0.5), "ValueError"
        )
        check = gp.check._check_type(subset_order, "subset_order", [int])
        if check:
            try:
                subset_order = int(subset_order)
                self._report(
                    gp.check._conversion(subset_order, "subset_order", int), "Warning"
                )
            except Exception:
                self._report(check, "TypeError")
        self._report(
            gp.check._check_value(subset_order, "subset_order", [1, 2]), "ValueError"
        )
        check = gp.check._check_type(adaptive_iterations, "adaptive_iterations", [int])
        if check:
            try:
                adaptive_iterations = int(adaptive_iterations)
                self._report(
                    gp.check._conversion(
                        adaptive_iterations, "adaptive_iterations", int
                    ),
                    "Warning",
                )
            except Exception:
                self._report(check, "TypeError")
        self._report(
            gp.check._check_range(adaptive_iterations, "adaptive_iterations", 0),
            "ValueError",
        )
        check = gp.check._check_type(correction, "correction", [bool])
        if check:
            try:
                correction = bool(correction)
                self._report(
                    gp.check._conversion(correction, "correction", int),
                    "Warning",
                )
            except Exception:
                self._report(check, "TypeError")
        self._report(
            gp.check._check_range(correction, "correction", 0),
            "ValueError",
        )
        check = gp.check._check_type(tolerance, "tolerance", [float])
        if check:
            try:
                tolerance = float(tolerance)
                self._report(
                    gp.check._conversion(tolerance, "tolerance", float), "Warning"
                )
            except Exception:
                self._report(check, "TypeError")
        self._report(
            gp.check._check_range(tolerance, "tolerance", 0.0, 1.0), "ValueError"
        )
        check = gp.check._check_type(seed_tolerance, "seed_tolerance", [float])
        if check:
            try:
                seed_tolerance = float(seed_tolerance)
                self._report(
                    gp.check._conversion(seed_tolerance, "seed_tolerance", float),
                    "Warning",
                )
            except Exception:
                self._report(check, "TypeError")
        self._report(
            gp.check._check_range(seed_tolerance, "seed_tolerance", 0.0, 1.0),
            "ValueError",
        )
        self._report(
            gp.check._check_comp(
                tolerance,
                "tolerance",
                seed_tolerance,
                "seed_tolerance",
            ),
            "ValueError",
        )
        self._report(gp.check._check_type(method, "method", [str]), "TypeError")
        self._report(
            gp.check._check_value(method, "method", ["FAGN", "ICGN"]), "ValueError"
        )
        check = gp.check._check_type(alpha, "alpha", [float])
        if check:
            try:
                alpha = float(alpha)
                self._report(gp.check._conversion(alpha, "alpha", float), "Warning")
            except Exception:
                self._report(check, "TypeError")
        self._report(gp.check._check_range(alpha, "alpha", 0.0, 1.0), "ValueError")

        # Store variables.
        self._seed_coord = seed_coord
        self._template = template
        self._max_iterations = max_iterations
        self._max_norm = max_norm
        self._adaptive_iterations = adaptive_iterations
        self._correction = correction
        self._method = method
        self._subset_order = subset_order
        self._tolerance = tolerance
        self._seed_tolerance = seed_tolerance
        self._alpha = alpha
        self._update = False
        self._seed_warp = seed_warp
        self._history = history
        self._override = override
        self._status = 0

        # Initialize gmsh if not already initialized.
        if gmsh.isInitialized() == 0:
            gmsh.initialize()
        gmsh.option.setNumber("General.Verbosity", 2)

        # Create initial mesh.
        self._initial_mesh()
        self._update_mesh()

        # Solve initial mesh.
        self._message = "Solving initial mesh"
        self._find_seed_node()
        try:
            self._reliability_guided()
            if self._unsolvable and not (self._status == 3 and self._override):
                self._check_status()
                return self.solved
            # Solve adaptive iterations.
            for iteration in range(1, adaptive_iterations + 1):
                self._message = "Adaptive iteration {}".format(iteration)
                self._adaptive_mesh()
                self._update_mesh()
                self._find_seed_node()
                self._reliability_guided()
                if self._unsolvable and not (self._status == 3 and self._override):
                    self._check_status()
                    return self.solved
            self._store_region()
            if self._unsolvable and not (self._status == 3 and self._override):
                self._check_status()
                return self.solved
            if self._status == 0:
                self._status = 1
            self.solved = True

            # Extract data from subsets.
            subset_data = []
            for subset in self._subsets:
                subset_data.append(subset.data)
            # Pack results.
            self._results = {
                "subsets": subset_data,
                "displacements": self._displacements,
                "warps": self._warps,
                "C_ZNCC": self._C_ZNCC,
                "seed": self._seed_node,
            }
            # Pack settings.
            self._settings = {
                "max_iterations": self._max_iterations,
                "max_norm": self._max_norm,
                "adaptive_iterations": self._adaptive_iterations,
                "method": self._method,
                "tolerance": self._tolerance,
                "seed_tolerance": self._seed_tolerance,
                "correction": self._correction,
                "alpha": self._alpha,
                "override": self._override,
            }
            # Pack data.
            self.data.update(
                {
                    "nodes": self._nodes,
                    "elements": self._elements,
                    "areas": self._areas,
                    "solved": self._solved,
                    "unsolvable": self._unsolvable,
                    "centroids": np.mean(self._nodes[self._elements], axis=-2),
                    "settings": self._settings,
                    "results": self._results,
                }
            )
            del self._history
        except Exception:
            self._status = 7
            self._unsolvable = True
            self._check_status()
            print(traceback.format_exc())
        gmsh.finalize()
        return self.solved

    def _check_status(self):
        """Private method for issue catergorisation and error message delivery."""
        if self._status == 0:  # Solvable unsolved.
            return
        elif self._status == 1:  # Solvable solved.
            log.info(
                "Solved mesh. Minimum correlation coefficient: {min_C:.2f}; "
                "maximum correlation coefficient: {max_C:.2f}.".format(
                    min_C=np.amin(self._C_ZNCC),
                    max_C=np.amax(self._C_ZNCC),
                )
            )
        elif self._status == 2:  # Unsolvable: seed decorrelation.
            log.error(
                "Specified seed correlation coefficient tolerance not met."
                "Minimum seed correlation coefficient:"
                "{seed_C:.2f}; tolerance: {seed_tolerance:.2f}.".format(
                    seed_C=self._subsets[self._seed_node].data["results"]["C_ZNCC"],
                    seed_tolerance=self._seed_tolerance,
                )
            )
        elif self._status == 3:  # Unsolvable: subset decorrelation.
            log.error(
                "Specified correlation coefficient tolerance not met. "
                "Minimum correlation coefficient: "
                "{min_C:.2f}; tolerance: {tolerance:.2f}.".format(
                    min_C=np.amin(self._C_ZNCC[np.where(self._C_ZNCC > 0.0)]),
                    tolerance=self._tolerance,
                )
            )
        elif self._status == 4:  # Unsolvable: rigid exclusions decorrelation.
            log.error("Rigid exclusion tracking failure.")
        elif self._status == 5:  # Unsolvable: unsolvable subset.
            log.error("Unsolvable subset.")
        elif self._status == 6:  # Unsolvable: subset compatability failure.
            log.error("Local mesh compatibility failure.")
        elif self._status == 7:  # Unsolvable: unkown issue.
            log.error("Could not solve mesh. Unrecognised problem.")
            input()
        self._update = True

    def _store_region(self):
        """Private method to record the deformation of the boundary and exclusions."""
        exclusion_obj_warp = []
        for i in range(np.shape(self._exclusion_objs)[0]):
            if self._exclusion_objs[i]._option == "R":
                subset = gp.subset.Subset(
                    f_img=self._f_img,
                    g_img=self._g_img,
                    f_coord=self._exclusion_objs[i]._centre,
                    template=gp.templates.Circle(
                        int(self._exclusion_objs[i]._specifics["radius"])
                    ),
                )
                warp_0 = np.zeros(6 * self._subset_order)
                warp_0[:2] = np.mean(self._displacements[self._exclusions[i]], axis=0)
                subset.solve(tolerance=0.9, warp_0=warp_0, order=self._subset_order)
                if subset.data["solved"] is not True:
                    self._status = 4
                    del subset
                    return
                else:
                    exclusion_obj_warp.append(subset._p.flatten())
                    del subset
            elif self._exclusion_objs[i]._option == "F":
                exclusion_obj_warp.append(self._displacements[self._exclusions[i]])
            else:
                exclusion_obj_warp.append(None)
        for i in range(np.shape(self._exclusion_objs)[0]):
            self._exclusion_objs[i]._store(exclusion_obj_warp[i])
        self._boundary_obj._store(self._displacements[self._boundary])

    def _update_region(self):
        """

        Private method to trigger region objects (boundary and exclusions) to update.

        """

        self._boundary_obj._update(self._f_img.filepath)
        for exclusion_obj in self._exclusion_objs:
            exclusion_obj._update(self._f_img.filepath)

    def _update_mesh(self):
        """

        Private method to update the mesh variables.

        """
        (
            _,
            nc,
            _,
        ) = gmsh.model.mesh.getNodes()  # Extracts: node coordinates.
        _, _, ent = gmsh.model.mesh.getElements(dim=2)  # Extracts: element node tags.
        self._nodes = np.column_stack(
            (nc[0::3], nc[1::3])
        )  # Nodal coordinate array (x,y).
        self._elements = np.reshape(
            (np.asarray(ent) - 1).flatten(), (-1, 3 * self._mesh_order)
        )  # Element connectivity array.
        self._boundary = gmsh.model.occ.getCurveLoops(0)[1][0]
        self._edges = list(gmsh.model.mesh.getNodesForPhysicalGroup(1, 0)[0] - 1)
        self._exclusions = []
        for i in range(len(self._exclusion_objs)):
            self._exclusions.append(np.sort(gmsh.model.occ.getCurveLoops(0)[1][i + 1]))
            self._edges += list(
                gmsh.model.mesh.getNodesForPhysicalGroup(1, i + 1)[0] - 1
            )

    def _find_seed_node(self):
        """

        Private method to find seed node given seed coordinate.

        """
        dist = np.sqrt(
            (self._nodes[:, 0] - self._seed_coord[0]) ** 2
            + (self._nodes[:, 1] - self._seed_coord[1]) ** 2
        )
        self._seed_node = np.argmin(dist)

    def _initial_mesh(self):
        """

        Private method to optimize the element size to generate
        approximately the desired number of elements.

        """

        def f(size):
            return self._uniform_remesh(
                size,
                self._borders,
                self._segments,
                self._curves,
                self._target_nodes,
                self._size_lower_bound,
                self._size_upper_bound,
                self._mesh_order,
            )

        minimize_scalar(
            f,
            bounds=(self._size_lower_bound, self._size_upper_bound),
            method="bounded",
        )

    def _adaptive_mesh(self):
        """

        Private method to perform adaptive remeshing.

        """
        message = "Adaptively remeshing..."
        with alive_bar(dual_line=True, bar=None, title=message) as bar:
            D = (
                abs(self._warps[:, 3] + self._warps[:, 4]) * self._areas
            )  # Elemental shear strain-area products.
            D_b = np.mean(D)  # Mean elemental shear strain-area product.
            self._areas *= (
                np.clip(D / D_b, self._alpha, 1 / self._alpha)
            ) ** -2  # Target element areas calculated.

            def f(scale):
                return self._adaptive_remesh(
                    scale,
                    self._target_nodes,
                    self._nodes,
                    self._elements,
                    self._areas,
                    self._mesh_order,
                )

            minimize_scalar(f, bounds=(0.1, 5))
            bar()

    @staticmethod
    def _uniform_remesh(
        size,
        boundary,
        segments,
        curves,
        target_nodes,
        size_lower_bound,
        size_upper_bound,
        order,
    ):
        """

        Private method to create the initial mesh.

        Parameters
        ----------
        size : int
            Target size of elements.
        boundary : `numpy.ndarray` (Nx,Ny)
            Array of coordinates to define the mesh boundary.
        segments : `numpy.ndarray` (Nx,Ny)
            Array of segments for gmsh mesh generation.
        curves : `numpy.ndarray` (Nx,Ny)
            Array of curves for gmsh mesh generation.
        target_nodes : int
            Target number of nodes.
        size_lower_bound : int
            Lower bound on element size.


        Returns
        -------
        error : int
            Error between target and actual number of nodes.

        """

        # Make mesh.
        gmsh.model.add("base")  # Create model.

        # Add points.
        for i in range(np.shape(boundary)[0]):
            gmsh.model.occ.addPoint(boundary[i, 0], boundary[i, 1], 0, size, i)

        # Add line segments.
        for i in range(np.shape(segments)[0]):
            gmsh.model.occ.addLine(segments[i, 0], segments[i, 1], i)

        # Add curves.
        for i in range(len(curves)):
            gmsh.model.occ.addCurveLoop(curves[i], i)
        curve_indices = list(np.arange(len(curves), dtype=np.int32))

        # Create surface.
        gmsh.model.occ.addPlaneSurface(curve_indices, 0)

        # Generate mesh.
        gmsh.option.setNumber("Mesh.MeshSizeMin", size_lower_bound)
        gmsh.option.setNumber("Mesh.MeshSizeMax", size_upper_bound)
        gmsh.model.occ.synchronize()

        # Add physical groups.
        for i in range(len(curves)):
            gmsh.model.addPhysicalGroup(1, curves[i], i)

        # Mesh.
        gmsh.model.mesh.generate(2)
        gmsh.model.mesh.optimize()
        gmsh.model.mesh.setOrder(order)

        # Get mesh topology.
        (
            _,
            nc,
            _,
        ) = (
            gmsh.model.mesh.getNodes()
        )  # Extracts: node tags, node coordinates, parametric coordinates.
        nodes = np.column_stack((nc[0::3], nc[1::3]))  # Nodal coordinate array (x,y).
        error = (np.shape(nodes)[0] - target_nodes) ** 2

        return error

    @staticmethod
    def _adaptive_remesh(scale, target, nodes, elements, areas, order):
        """

        Private method to perform adaptive mesh generation.

        Parameters
        ----------
        scale : float
            Scale factor on element size.
        target : int
            Target number of nodes.
        nodes : `numpy.ndarray`
            Nodes for background mesh.
        elements : `numpy.ndarray`
            Elements for background mesh.
        areas : float
            Target element areas.


        Returns
        -------
        error : float
            Error between target and actual number of nodes.

        """
        lengths = gp.geometry.utilities.area_to_length(
            areas * scale
        )  # Convert target areas to target characteristic lengths.
        bg = gmsh.view.add("bg", 1)  # Create background view.
        data = np.pad(
            nodes[elements[:, :3]],
            ((0, 0), (0, 0), (0, 2)),
            mode="constant",
        )  # Prepare data input (coordinates and buffer).

        data[:, :, 3] = np.reshape(
            np.repeat(lengths, 3), (-1, 3)
        )  # Fill data input buffer with target weights.
        data = np.transpose(data, (0, 2, 1)).flatten()  # Reshape for input.
        gmsh.view.addListData(bg, "ST", len(elements), data)  # Add data to view.
        bgf = gmsh.model.mesh.field.add("PostView")  # Add view to field.
        gmsh.model.mesh.field.setNumber(
            bgf, "ViewTag", bg
        )  # Establish field reference (important for function reuse).
        gmsh.model.mesh.field.setAsBackgroundMesh(bgf)  # Set field as background.
        gmsh.option.setNumber(
            "Mesh.MeshSizeExtendFromBoundary", 0
        )  # Prevent boundary influence on mesh.
        gmsh.option.setNumber(
            "Mesh.MeshSizeFromPoints", 0
        )  # Prevent point influence on mesh.
        gmsh.option.setNumber(
            "Mesh.MeshSizeFromCurvature", 0
        )  # Prevent curve influence on mesh.
        gmsh.option.setNumber("Mesh.Algorithm", 5)
        gmsh.model.mesh.clear()  # Tidy.
        gmsh.model.mesh.generate(2)  # Generate mesh.
        gmsh.model.mesh.optimize()
        gmsh.model.mesh.setOrder(order)

        (
            nt,
            nc,
            npar,
        ) = (
            gmsh.model.mesh.getNodes()
        )  # Extracts: node tags, node coordinates, parametric coordinates.
        nodes = np.column_stack((nc[0::3], nc[1::3]))  # Nodal coordinate array (x,y).
        error = (np.shape(nodes)[0] - target) ** 2
        return error

    def _element_area(self):
        """
        Private method to calculate the element areas based on corner nodes.

        """
        M = np.ones((len(self._elements), 3, 3))
        M[:, 1] = self._nodes[self._elements[:, :3]][:, :, 0]
        M[:, 2] = self._nodes[self._elements[:, :3]][:, :, 1]
        self._areas = 0.5 * np.linalg.det(M)

    def _local_coordinates(self):
        """
        Private method to define the element centroid in terms of the local
        coordinate system. Thereotically, this is constant but is calculated in
        case of rounding error.

        Return
        ------
        A : `numpy.ndarray` (N,3,4)
            Element centroid local coordinates.
        """

        A = np.ones((np.shape(self._elements)[0], 3, 4))
        A[:, 1:, 0] = np.mean(self._nodes[self._elements], axis=1)
        A[:, 1:, 1:] = self._nodes[self._elements][:, :3, :2].transpose(0, 2, 1)

        return A

    def _shape_function(self):
        """Private method for defining the shape functions at the
        element centroid.

        Return
        ------
        N : `numpy.ndarray`
            Shape function.
        dN : `numpy.ndarray`
            Shape function 1st order derivative.
        d2N : `numpy.ndarray`
            Shape function 2nd order derivative.
        """

        if self._mesh_order == 1:
            N = np.asarray([1 / 3, 1 / 3, 1 / 3])
            dN = np.asarray([[1, 0, -1], [0, 1, -1]])
            d2N = None
        elif self._mesh_order == 2:
            N = np.asarray([-1 / 9, -1 / 9, -1 / 9, 4 / 9, 4 / 9, 4 / 9])
            dN = np.asarray(
                [
                    [1 / 3, 0, -1 / 3, 4 / 3, -4 / 3, 0],
                    [0, 1 / 3, -1 / 3, 4 / 3, 0, -4 / 3],
                ]
            )
            d2N = np.asarray(
                [
                    [4, 0, 4, 0, 0, -8],
                    [0, 0, 4, 4, -4, -4],
                    [0, 4, 4, 0, -8, 0],
                ]
            )
        return N, dN, d2N

    def _element_strains(self):
        """

        Private method to calculate the elemental strain the "B" matrix
        relating element node displacements to elemental strain.

        """

        # Setup.
        self._warps = np.zeros((np.shape(self._elements)[0], 12))
        x = self._nodes[self._elements]
        u = self._displacements[self._elements]

        # Coordinate matrix.
        A = self._local_coordinates()

        # Shape function and derivatives.
        N, dN, d2N = self._shape_function()

        # Displacements.
        self._warps[:, :2] = N @ u

        # 1st Order Strains
        J_x_T = dN @ x
        J_u_T = dN @ u

        self._warps[:, 2:6] = (np.linalg.inv(J_x_T) @ J_u_T).reshape(
            np.shape(self._elements)[0], -1
        )

        # 2nd Order Strains
        if self._mesh_order == 2:
            K_u = d2N @ u
            dz = np.zeros((np.shape(self._elements)[0], 2, 2))
            dz[:, 0, 0] = x[:, 1, 1] - x[:, 2, 1]
            dz[:, 0, 1] = x[:, 2, 1] - x[:, 0, 1]
            dz[:, 1, 0] = x[:, 2, 0] - x[:, 1, 0]
            dz[:, 1, 1] = x[:, 0, 0] - x[:, 2, 0]
            dz /= np.linalg.det(A[:, :, [1, 2, 3]])[:, None, None]

            K_x_inv = np.zeros((np.shape(self._elements)[0], 3, 3))
            K_x_inv[:, 0, 0] = dz[:, 0, 0] ** 2
            K_x_inv[:, 0, 1] = 2 * dz[:, 0, 0] * dz[:, 0, 1]
            K_x_inv[:, 0, 2] = dz[:, 0, 1] ** 2
            K_x_inv[:, 1, 0] = dz[:, 0, 0] * dz[:, 1, 0]
            K_x_inv[:, 1, 1] = dz[:, 0, 0] * dz[:, 1, 1] + dz[:, 0, 1] * dz[:, 1, 0]
            K_x_inv[:, 1, 2] = dz[:, 0, 1] * dz[:, 1, 1]
            K_x_inv[:, 2, 0] = dz[:, 1, 0] ** 2
            K_x_inv[:, 2, 1] = 2 * dz[:, 1, 0] * dz[:, 1, 1]
            K_x_inv[:, 2, 2] = dz[:, 1, 1] ** 2

            self._warps[:, 6:] = (K_x_inv @ K_u).reshape(
                np.shape(self._elements)[0], -1
            )

    def _seed_solve(self):
        """

        Private method for the solving of the seed node.

        """

        self._subsets[self._seed_node].solve(
            max_norm=self._max_norm,
            max_iterations=self._max_iterations,
            warp_0=self._seed_warp,
            order=self._subset_order,
            method=self._method,
            tolerance=self._seed_tolerance,
        )  # Solve for seed subset.
        if not self._subsets[self._seed_node].data["solved"]:  # If seed unsolved.
            self._subsets[self._seed_node].solve(
                max_norm=self._max_norm,
                max_iterations=self._max_iterations,
                warp_0=np.zeros(np.shape(self._seed_warp)),
                order=self._subset_order,
                method=self._method,
                tolerance=self._seed_tolerance,
            )
            if not self._subsets[self._seed_node].data["solved"]:  # If seed unsolved.
                self._unsolvable = True
                self._status = 2
                return
        # Store if solved.
        self._store_variables(idx=self._seed_node, flag=-1)

    def _output_preparation(self):
        # Set up.
        m = np.shape(self._nodes)[0]
        n = np.shape(self._seed_warp)[0]
        self._subset_solved = np.zeros(
            m, dtype=int
        )  # Solved/unsolved reference array (1 if unsolved, -1 if solved).
        self._C_ZNCC = np.zeros(m, dtype=np.float64)  # Correlation coefficient array.
        self._subsets = np.empty(m, dtype=object)  # Initiate subset array.
        self._p = np.zeros((m, n), dtype=np.float64)  # Warp function array.
        self._displacements = np.zeros(
            (m, 2), dtype=np.float64
        )  # Displacement output array.

    def _reliability_guided(self):
        """

        Private method to perform reliability-guided (RG) PIV analysis.

        """

        self._output_preparation()

        # Subset instantiation with template masking.
        self._subset_instantiation()

        # Solve subsets in mesh.
        with alive_bar(
            np.shape(self._nodes)[0],
            dual_line=True,
            bar="blocks",
            title=self._message,
        ) as self._bar:
            # Solve for seed.
            self._bar.text = "-> Solving seed subset..."
            self._seed_solve()
            if self._unsolvable and not (self._status == 3 and self._override):
                return
            self._bar()

            # Solve for seed neighbours.
            self._neighbours(
                self._seed_node, self._subsets[self._seed_node].data["results"]["p"]
            )
            if self._unsolvable and not (self._status == 3 and self._override):
                return
            # Solve through sorted queue.
            self._bar.text = (
                "-> Solving remaining subsets using reliability guided approach..."
            )
            while np.max(self._subset_solved) > -1:
                # Highest correlation subset selected.
                cur_idx = np.argmax(self._subset_solved * self._C_ZNCC)
                self._subset_solved[cur_idx] = -1  # Set as solved.
                p_0 = self._subsets[cur_idx].data["results"]["p"]  # Precondition.
                self._neighbours(cur_idx, p_0)  # Solve for neighbours.
                self._bar()
                if self._bar.current == np.shape(self._nodes)[0]:
                    break
                if self._unsolvable and not (self._status == 3 and self._override):
                    return
        # Corrections, calculations and checks.
        if self._correction:
            self._corrections()
        # self._compatibility()
        if self._unsolvable and not (self._status == 3 and self._override):
            return
        self._element_area()
        self._element_strains()
        if any(self._subset_solved != -1):
            self._unsolvable = True
            self._status = 5
        elif any(self._C_ZNCC < self._tolerance):
            self._unsolvable = True
            self._status = 3

    def _subset_instantiation(self):
        """Private method that instantiates the mesh subsets with masking."""

        for i in range(np.shape(self._nodes)[0]):
            if i in self._edges:
                template = deepcopy(self._template)
                template.mask(self._nodes[i], self._mask)
                check = False
                for e in range(len(self._exclusions)):
                    if i in self._exclusions[e]:
                        if self._exclusion_objs[e]._compensate:
                            check = True
                            break
                if (
                    i in self._boundary and self._boundary_obj._compensate
                ) or check is True:
                    if template.n_px < self._template.n_px:
                        size = int(
                            self._template.size
                            / np.sqrt(template.n_px / self._template.n_px)
                        )
                        if self._template.shape == "circle":
                            template = gp.templates.Circle(radius=size)
                        elif self._template.shape == "square":
                            template = gp.templates.Square(length=size)
                template.mask(self._nodes[i], self._mask)
                self._subsets[i] = gp.subset.Subset(
                    f_coord=self._nodes[i],
                    f_img=self._f_img,
                    g_img=self._g_img,
                    template=template,
                )  # Create masked boundary subset.
            else:
                self._subsets[i] = gp.subset.Subset(
                    f_coord=self._nodes[i],
                    f_img=self._f_img,
                    g_img=self._g_img,
                    template=self._template,
                )  # Create full subset.

    def _compatibility(self):
        """

        Private method to ensure compatibility (that the displacement of a subset
        nowhere exceeds the displacement bounds of neighbouring subsets).

        """
        for i in range(np.shape(self._subsets)[0]):
            if i not in self._edges:
                neighbours = self._connectivity(i, full=True)
                if len(neighbours) > 2:
                    hull = sp.spatial.Delaunay(
                        self._nodes[neighbours] + self._displacements[neighbours]
                    )
                    if hull.find_simplex(self._nodes[i] + self._displacements[i]) < 0:
                        fixed = self._compatability_correction(i, neighbours, hull)
                        if fixed is not True:
                            self._unsolvable = True
                            self._status = 6

    def _compatability_correction(self, idx, neighbours, hull):
        """

        Private method to correct subset displacement in the case of incompatibility.

        """

        subset = gp.subset.Subset(
            f_coord=self._nodes[idx],
            f_img=self._f_img,
            g_img=self._g_img,
            template=self._subsets[idx]._template,
        )
        warp = np.mean(
            self._p[neighbours[self._C_ZNCC[neighbours] > self._tolerance]],
            axis=0,
        )
        subset.solve(
            max_norm=self._max_norm,
            max_iterations=self._max_iterations,
            warp_0=warp,
            order=self._subset_order,
            method=self._method,
            tolerance=self._seed_tolerance,
        )
        coord = self._nodes[idx] + subset._p[:2].flatten()
        if hull.find_simplex(coord) >= 0 and subset.solved:
            self._subsets[idx] = subset
            self._store_variables(idx=idx, flag=-1)
            return True
        else:
            return False

    def _corrections(self):
        """

        Private method to improve subset correlation through neighbour
        preconditioning.

        """

        arg = np.argsort(self._C_ZNCC)
        corarg = arg[self._C_ZNCC[arg] < self._seed_tolerance]
        j = 0
        k = 0
        benefit = np.zeros(len(self._C_ZNCC))
        if np.shape(corarg)[0] >= 1:
            with alive_bar(
                np.shape(corarg)[0],
                dual_line=True,
                bar="blocks",
                title="Corrections",
            ) as bar:
                for idx in corarg:
                    k += 1
                    subset = gp.subset.Subset(
                        f_coord=self._nodes[idx],
                        f_img=self._f_img,
                        g_img=self._g_img,
                        template=self._subsets[idx]._template,
                    )
                    neighbours = self._connectivity(idx)
                    warp = np.mean(
                        self._p[neighbours[self._C_ZNCC[neighbours] > self._tolerance]],
                        axis=0,
                    )
                    subset.solve(
                        max_norm=self._max_norm,
                        max_iterations=self._max_iterations,
                        warp_0=warp,
                        order=self._subset_order,
                        method=self._method,
                        tolerance=self._seed_tolerance,
                    )
                    try:
                        if (
                            subset._C_ZNCC > self._subsets[idx]._C_ZNCC
                            and subset.solved
                        ):
                            j += 1
                            benefit[idx] = subset._C_ZNCC - self._subsets[idx]._C_ZNCC
                            self._subsets[idx] = subset
                            self._store_variables(idx=idx, flag=-1)
                    except Exception:
                        pass
                    bar()
                    bar.text = "{} corrections accepted".format(round(j / k, 3))
            log.info(
                (
                    "{ca} corrections accepted;"
                    "{max} maximum improvement;"
                    "{mean} mean improvement"
                ).format(
                    ca=round(j / k, 3),
                    max=round(np.max(benefit), 3),
                    mean=round(np.mean(benefit), 3),
                )
            )

    def _connectivity(self, idx, full=False):
        """

        A private method that returns the indices of nodes connected
        to the index node according to the input array.

        Parameters
        ----------
        idx : int
            Index of node.
        full : bool, optional
            False - return immediate nodes to the index.
            True - return all nodes sharing an element with the index.


        Returns
        -------
        pts_idx : `numpy.ndarray` (N)
            Mesh array.

        """

        if self._mesh_order == 1 or full is True:
            element_idxs = np.argwhere(np.any(self._elements == idx, axis=1)).flatten()
            pts_idxs = np.unique(self._elements[element_idxs])
            pts_idxs = np.delete(pts_idxs, np.argwhere(pts_idxs == idx))
        elif self._mesh_order == 2:
            element_idxs = np.argwhere(self._elements == idx)
            pts_idxs = []
            for i in range(len(element_idxs)):
                if element_idxs[i, 1] == 0:  # If 1
                    pts_idxs.append(self._elements[element_idxs[i, 0], 3::2])  # Add 4,6
                elif element_idxs[i, 1] == 1:  # If 2
                    pts_idxs.append(self._elements[element_idxs[i, 0], 3:5])  # Add 4,5
                elif element_idxs[i, 1] == 2:  # If 3
                    pts_idxs.append(self._elements[element_idxs[i, 0], 4:])  # Add 5,6
                elif element_idxs[i, 1] == 3:  # If 4
                    pts_idxs.append(self._elements[element_idxs[i, 0], :2])  # Add 1,2
                elif element_idxs[i, 1] == 4:  # If 5
                    pts_idxs.append(self._elements[element_idxs[i, 0], 1:3])  # Add 2,3
                elif element_idxs[i, 1] == 5:  # If 6
                    pts_idxs.append(self._elements[element_idxs[i, 0], :3:2])  # Add 1,3
            pts_idxs = np.unique(pts_idxs)

        return pts_idxs

    def _neighbours(self, cur_idx, p_0):
        """

        Private method to calculate the correlation coefficients and
        warp functions of the neighbouring nodes.

        Parameters
        ----------
        cur_idx : int
            The current susbet index.
        p_0 : `numpy.ndarray` (N)
            Preconditioning warp function.


        Returns
        -------
        solved : bool
            Boolean to indicate whether the neighbouring subsets have been solved.

        """

        # Identify neighbours.
        neighbours = self._connectivity(cur_idx)

        # Iterate through list of neighbours.
        for idx in neighbours:
            if self._subset_solved[idx] == 0:  # If not previously solved.
                # if self._history:
                #     # 0. Use history pre-conditioning
                #     particle = gp.particle.Particle(
                #             series = self._history,
                #             coordinate_0 = self._nodes[idx]
                #         )
                #     try:
                #         particle.solve()
                #         p_h = particle._warps[-1]
                #         self._subsets[idx].solve(
                #             max_norm=self._max_norm,
                #             max_iterations=self._max_iterations,
                #             order=self._subset_order,
                #             warp_0=p_h,
                #             method=self._method,
                #             tolerance=self._tolerance,
                #         )
                #     except Exception:
                #         pass
                #     del particle
                # if self._subsets[idx].data["solved"]:  # Check against tolerance.
                #    self._store_variables(idx=idx, flag=1)
                # elif self._subsets[idx].data["unsolvable"]:
                #    self._status = 5
                #    self._unsolvable = True
                # else:
                # 1. Use nearest-neighbour pre-conditioning.
                self._subsets[idx].solve(
                    max_norm=self._max_norm,
                    max_iterations=self._max_iterations,
                    order=self._subset_order,
                    warp_0=p_0,
                    method=self._method,
                    tolerance=self._tolerance,
                )
                if self._subsets[idx].data["solved"]:  # Check against tolerance.
                    self._store_variables(idx=idx, flag=1)
                elif self._subsets[idx].data["unsolvable"]:
                    self._status = 5
                    self._unsolvable = True
                else:
                    # 2. Use projected pre-conditioning.
                    diff = self._nodes[idx] - self._nodes[cur_idx]
                    p = self._subsets[cur_idx].data["results"]["p"]
                    if np.shape(p)[0] == 6:
                        p_0[0] = p[0] + p[2] * diff[0] + p[3] * diff[1]
                        p_0[1] = p[1] + p[4] * diff[0] + p[5] * diff[1]
                    elif np.shape(p)[0] == 12:
                        p_0[0] = (
                            p[0]
                            + p[2] * diff[0]
                            + p[4] * diff[1]
                            + 0.5 * p[6] * diff[0] ** 2
                            + p[8] * diff[0] * diff[1]
                            + 0.5 * p[10] * diff[1] ** 2
                        )
                        p_0[1] = (
                            p[1]
                            + p[3] * diff[0]
                            + p[5] * diff[1]
                            + 0.5 * p[7] * diff[0] ** 2
                            + p[9] * diff[0] * diff[1]
                            + 0.5 * p[11] * diff[1] ** 2
                        )
                        p_0[2] = p[2] + p[6] * diff[0] + p[8] * diff[1]
                        p_0[3] = p[3] + p[7] * diff[0] + p[9] * diff[1]
                        p_0[4] = p[4] + p[8] * diff[1] + p[10] * diff[1]
                        p_0[5] = p[5] + p[9] * diff[1] + p[11] * diff[1]
                    self._subsets[idx].solve(
                        max_norm=self._max_norm,
                        max_iterations=self._max_iterations,
                        warp_0=p_0,
                        method=self._method,
                        tolerance=self._tolerance,
                        order=self._subset_order,
                    )
                    if self._subsets[idx].data["solved"]:  # Check against tolerance.
                        self._store_variables(idx=idx, flag=1)
                    elif self._subsets[idx].data["unsolvable"]:
                        self._status = 5
                        self._unsolvable = True
                    else:
                        # 3. Use the NCC initial guess.
                        self._subsets[idx].solve(
                            max_norm=self._max_norm,
                            max_iterations=self._max_iterations,
                            warp_0=np.zeros(np.shape(p_0)),
                            method=self._method,
                            tolerance=self._tolerance,
                            order=self._subset_order,
                        )
                        if self._subsets[idx].data["unsolvable"] or np.isnan(
                            self._subsets[idx].data["results"]["C_ZNCC"]
                        ):
                            self._status = 5
                            self._unsolvable = True
                        else:
                            self._store_variables(idx=idx, flag=1)

    def _store_variables(self, idx, flag):
        """

        Private method to store variables.

        Parameters
        ----------
        idx : int
            Index of current subset.
        flag : int
            -1 : solved and stored
            0  : unsolved
            1  : solved and unstored

        """

        self._subset_solved[idx] = flag
        self._C_ZNCC[idx] = np.max(
            (self._subsets[idx].data["results"]["C_ZNCC"], 0)
        )  # Clip correlation coefficient to positive values.
        self._p[idx] = self._subsets[idx].data["results"]["p"].flatten()
        self._displacements[idx, 0] = self._subsets[idx].data["results"]["u"]
        self._displacements[idx, 1] = self._subsets[idx].data["results"]["v"]


class MeshResults(MeshBase):
    """

    MeshResults class for geopyv.

    Parameters
    ----------
    data : dict
        geopyv data dict from Mesh object.


    Attributes
    ----------
    data : dict
        geopyv data dict from Mesh object.

    """

    def __init__(self, data):
        """Initialisation of geopyv MeshResults class."""
        self.data = data



"""

Object module for geopyv.

"""


class Object:
    """

    Base class for any geopyv object.

    """

    def __init__(self, object_type):
        """

        Base class object initialiser.

        Parameters
        ----------
        object_type : str
            Object type.

        """
        self.object_type = object_type



"""

Calibration module for geopyv.

"""

import logging
import numpy as np
import geopyv as gp
from geopyv.object import Object
import cv2
from cv2 import aruco
import matplotlib.pyplot as plt
import glob
import re
from scipy.optimize import root
from alive_progress import alive_bar
import traceback

log = logging.getLogger(__name__)


class CalibrationBase(Object):
    """

    Calibration base class to be used as a mixin. Contains plot functionality.

    """

    def __init__(self):
        super().__init__(object_type="Calibration")
        """

        Calibration base class initialiser.

        """

    def inspect(self, *, image_index=0):
        fig, ax = gp.plots.inspect_calibration(data=self.data, image_index=image_index)

    def visualise(
        self,
        *,
        show=True,
        block=True,
        save=None,
    ):
        # Check if solved.
        if self.data["solved"] is not True:
            log.error(
                "Calibration not yet solved therefore nothing to inspect. "
                "First, run :meth:`~geopyv.calibration.Calibration.solve()` to solve."
            )
            raise ValueError(
                "Calibration not yet solved therefore nothing to inspect. "
                "First, run :meth:`~geopyv.calibration.Calibration.solve()` to solve."
            )
        self._report(gp.check._check_type(show, "show", [bool]), "TypeError")
        self._report(gp.check._check_type(block, "block", [bool]), "TypeError")
        self._report(gp.check._check_type(save, "save", [str, type(None)]), "TypeError")

        gp.plots.visualise_calibration(self.data, block, show, save)

    def contour(
        self,
        *,
        quantity="R",
        points=True,
        colorbar=True,
        ticks=None,
        alpha=0.75,
        levels=None,
        axis=True,
        xlim=None,
        ylim=None,
        show=True,
        block=True,
        save=None,
    ):
        # Check if solved.
        if self.data["solved"] is not True:
            log.error(
                "Calibration not yet solved therefore nothing to inspect. "
                "First, run :meth:`~geopyv.calibration.Calibration.solve()` to solve."
            )
            raise ValueError(
                "Calibration not yet solved therefore nothing to inspect. "
                "First, run :meth:`~geopyv.calibration.Calibration.solve()` to solve."
            )
        types = [
            "u",
            "v",
            "R",
        ]
        if quantity:
            self._report(
                gp.check._check_value(quantity, "quantity", types), "ValueError"
            )
        self._report(gp.check._check_type(colorbar, "colorbar", [bool]), "TypeError")
        types = [tuple, list, np.ndarray, type(None)]
        self._report(gp.check._check_type(ticks, "ticks", types), "TypeError")
        check = gp.check._check_type(alpha, "alpha", [float])
        if check:
            try:
                alpha = float(alpha)
                self._report(gp.check._conversion(alpha, "alpha", float), "Warning")
            except Exception:
                self._report(check, "TypeError")
        self._report(gp.check._check_range(alpha, "alpha", 0.0, 1.0), "ValueError")
        self._report(gp.check._check_type(levels, "levels", types), "TypeError")
        self._report(gp.check._check_type(axis, "axis", [bool]), "TypeError")
        types = [tuple, list, np.ndarray, type(None)]
        self._report(gp.check._check_type(xlim, "xlim", types), "TypeError")
        if xlim is not None:
            self._report(gp.check._check_dim(xlim, "xlim", 1), "ValueError")
            self._report(gp.check._check_axis(xlim, "xlim", 0, [2]), "ValueError")
        self._report(gp.check._check_type(ylim, "ylim", types), "TypeError")
        if ylim is not None:
            self._report(gp.check._check_dim(ylim, "ylim", 1), "ValueError")
            self._report(gp.check._check_axis(ylim, "ylim", 0, [2]), "ValueError")
        self._report(gp.check._check_type(show, "show", [bool]), "TypeError")
        self._report(gp.check._check_type(block, "block", [bool]), "TypeError")
        self._report(gp.check._check_type(save, "save", [str, type(None)]), "TypeError")

        gp.plots.contour_calibration(
            self.data,
            quantity,
            points,
            colorbar,
            ticks,
            alpha,
            levels,
            axis,
            xlim,
            ylim,
            block,
            show,
            save,
        )

    def error(
        self,
        *,
        quantity="R",
        points=True,
        colorbar=True,
        ticks=None,
        alpha=0.75,
        levels=None,
        axis=True,
        xlim=None,
        ylim=None,
        show=True,
        block=True,
        save=None,
    ):
        # Check if solved.
        if self.data["solved"] is not True:
            log.error(
                "Calibration not yet solved therefore nothing to inspect. "
                "First, run :meth:`~geopyv.calibration.Calibration.solve()` to solve."
            )
            raise ValueError(
                "Calibration not yet solved therefore nothing to inspect. "
                "First, run :meth:`~geopyv.calibration.Calibration.solve()` to solve."
            )
        types = [
            "u",
            "v",
            "R",
        ]
        if quantity:
            self._report(
                gp.check._check_value(quantity, "quantity", types), "ValueError"
            )
        self._report(gp.check._check_type(points, "points", [bool]), "TypeError")
        self._report(gp.check._check_type(colorbar, "colorbar", [bool]), "TypeError")
        types = [tuple, list, np.ndarray, type(None)]
        self._report(gp.check._check_type(ticks, "ticks", types), "TypeError")
        check = gp.check._check_type(alpha, "alpha", [float])
        if check:
            try:
                alpha = float(alpha)
                self._report(gp.check._conversion(alpha, "alpha", float), "Warning")
            except Exception:
                self._report(check, "TypeError")
        self._report(gp.check._check_range(alpha, "alpha", 0.0, 1.0), "ValueError")
        self._report(gp.check._check_type(levels, "levels", types), "TypeError")
        self._report(gp.check._check_type(axis, "axis", [bool]), "TypeError")
        types = [tuple, list, np.ndarray, type(None)]
        self._report(gp.check._check_type(xlim, "xlim", types), "TypeError")
        if xlim is not None:
            self._report(gp.check._check_dim(xlim, "xlim", 1), "ValueError")
            self._report(gp.check._check_axis(xlim, "xlim", 0, [2]), "ValueError")
        self._report(gp.check._check_type(ylim, "ylim", types), "TypeError")
        if ylim is not None:
            self._report(gp.check._check_dim(ylim, "ylim", 1), "ValueError")
            self._report(gp.check._check_axis(ylim, "ylim", 0, [2]), "ValueError")
        self._report(gp.check._check_type(show, "show", [bool]), "TypeError")
        self._report(gp.check._check_type(block, "block", [bool]), "TypeError")
        self._report(gp.check._check_type(save, "save", [str, type(None)]), "TypeError")

        gp.plots.error_calibration(
            self.data,
            quantity,
            points,
            colorbar,
            ticks,
            alpha,
            levels,
            axis,
            xlim,
            ylim,
            block,
            show,
            save,
        )

    def o2i(self, objpnts):
        """Method for mapping coordinates from object space to image space."""

        # Checks for format of objpnts (should be [x,y,0,1]).
        if np.shape(objpnts)[-1] == 2:
            objpnts = np.pad(objpnts, (0, 1))[:-1]
        if np.shape(objpnts)[-1] == 3:
            objpnts = np.pad(objpnts, (0, 1), constant_values=(0, 1))[:-1]

        # Mapping.
        X_c = self._extmat @ objpnts.T
        X_c /= X_c[2]
        r2 = X_c[0] ** 2 + X_c[1] ** 2
        f = 1 + self._dist[0] * r2 + self._dist[1] * r2**2 + self._dist[4] * r2**3
        X_pp = np.ones((np.shape(objpnts)[0], 3))
        X_pp[:, 0] = (
            X_c[0] * f
            + 2 * self._dist[2] * X_c[0] * X_c[1]
            + self._dist[3] * (r2 + 2 * X_c[0] ** 2)
        )
        X_pp[:, 1] = (
            X_c[1] * f
            + self._dist[2] * (r2 + 2 * X_c[1] ** 2)
            + 2 * self._dist[3] * X_c[0] * X_c[1]
        )

        return (self._intmat @ X_pp.T).T[:, :2]

    def i2o(self, *, imgpnts, both=False):
        # Checks for format of imgpnts as in o2i.
        if np.ndim(imgpnts) <= 1:
            imgpnts = imgpnts[np.newaxis, ...]
        if np.shape(imgpnts)[-1] == 2:
            imgpnts = np.pad(imgpnts, (0, 1), constant_values=(0, 1))[:-1]

        # Mapping.
        X_pp = np.linalg.inv(self._intmat) @ imgpnts.T
        X_c = np.pad(self._simult(X_pp), (0, 1), constant_values=(0, 1))[:, :-1]
        X_c = self._depth_recovery(X_c)
        X_c = np.pad(X_c, (0, 1), constant_values=(0, 1))[:, :-1]
        objpnts = (np.linalg.inv(self._extmat) @ X_c).T[:, :3]
        if both is True:
            return objpnts, imgpnts
        else:
            return objpnts

    def _depth_recovery(self, X_c):
        invextmat = np.linalg.inv(self._extmat)
        X_c *= -invextmat[2, 3] / (
            invextmat[2, 0] * X_c[0] + invextmat[2, 1] * X_c[1] + invextmat[2, 2]
        )

        return X_c

    def _simult(self, X_pp):
        solution = root(
            lambda X_c: self._equations(X_c, X_pp), np.zeros((2, np.shape(X_pp)[-1]))
        )

        return solution.x.reshape(2, -1)

    def _equations(self, X_c, X_pp):
        X_c = np.reshape(X_c, (2, -1))
        r2 = np.sum(np.square(X_c), axis=0)  # X_c[0] ** 2 + X_c[1] ** 2
        f = 1 + self._dist[0] * r2 + self._dist[1] * r2**2 + self._dist[4] * r2**3
        # a = (
        #     X_c[0] * f
        #     + 2 * self._dist[2] * X_c[0] * X_c[1]
        #     + self._dist[3] * (r2 + 2 * X_c[0]**2)
        #     - X_pp[0]
        # )
        # b = (
        #     X_c[1] * f
        #     + 2 * self._dist[3] * X_c[0] * X_c[1]
        #     + self._dist[2] * (r2 + 2 * X_c[1]**2)
        #     - X_pp[1]
        # )
        # return np.asarray([a,b]).flatten()

        c = (
            X_c * f
            + 2 * self._dist[2:4, None] * (X_c[0] * X_c[1])
            + self._dist[3:1:-1, None] * (r2 + 2 * X_c**2)
            - X_pp[:2]
        )
        return c.flatten()

    def calibrate(self, *, opt=1, verbose=True, object=None):
        if self.data["solved"] is not True:
            log.error(
                "Calibration not yet solved therefore nothing to inspect. "
                "First, run :meth:`~geopyv.calibration.Calibration.solve()` to solve."
            )
            raise ValueError(
                "Calibration not yet solved therefore nothing to inspect. "
                "First, run :meth:`~geopyv.calibration.Calibration.solve()` to solve."
            )
        self._report(
            gp.check._check_type(
                object,
                "object",
                [
                    gp.sequence.Sequence,
                    gp.sequence.SequenceResults,
                    gp.mesh.Mesh,
                    gp.mesh.MeshResults,
                    gp.subset.Subset,
                    gp.subset.SubsetResults,
                    dict,
                ],
            ),
            "TypeError",
        )

        if type(object) == dict:
            data = object
        else:
            data = object.data

        if data["type"] == "Subset":
            if verbose:
                log.info("Calibrating Subset ...")
            # Calibrating.
            X, x = self.i2o(
                imgpnts=np.asarray(
                    [
                        [
                            data["position"]["x"],
                            data["position"]["y"],
                        ],
                        [
                            data["position"]["x"] + data["results"]["u"],
                            data["position"]["y"] + data["results"]["v"],
                        ],
                    ]
                ),
                both=True,
            )
            # Storing.
            data["position"].update(
                {"X": X[0, 0], "Y": X[0, 1], "x": x[0, 0], "y": x[0, 1]}
            )
            data["results"].update(
                {
                    "U": X[1, 0] - X[0, 0],
                    "V": X[1, 1] - X[0, 1],
                    "u": x[1, 0] - x[0, 0],
                    "v": x[1, 1] - x[0, 1],
                }
            )
            data["calibrated"] = True
            # Assigning.
            if type(object) == gp.subset.Subset:
                object._X = X[0, 0]
                object._Y = X[0, 1]
                object._U = X[1, 0] - X[0, 0]
                object._V = X[1, 1] - X[0, 1]
                object._x = x[0, 0]
                object._y = x[0, 1]
                object._u = x[1, 0] - x[0, 0]
                object._v = x[1, 1] - x[0, 1]
                object._calibrated = True
        elif data["type"] == "Mesh":
            if verbose:
                log.info("Calibrating Mesh ...")
            # Calibrating.
            Nodes, nodes = self.i2o(
                imgpnts=data["nodes"],
                both=True,
            )
            Nodes = Nodes[:, :2]
            nodes = nodes[:, :2]
            centroids = np.mean(nodes[data["elements"]], axis=-2)
            Centroids = np.mean(Nodes[data["elements"]], axis=-2)
            Displacements, displacements = self.i2o(
                imgpnts=(data["nodes"] + data["results"]["displacements"]), both=True
            )
            displacements = displacements[:, :2] - nodes
            Displacements = Displacements[:, :2] - Nodes
            # Distribute.
            for i in range(len(data["results"]["subsets"])):
                s = data["results"]["subsets"][i]
                s["position"].update(
                    {
                        "X": Nodes[i, 0],
                        "Y": Nodes[i, 1],
                        "x": nodes[i, 0],
                        "y": nodes[i, 1],
                    }
                )
                s["results"].update(
                    {
                        "U": Displacements[i, 0],
                        "V": Displacements[i, 1],
                        "u": displacements[i, 0],
                        "v": displacements[i, 1],
                    }
                )
                s["calibrated"] = True
            # Storing.
            data.update(
                {
                    "calibrated": True,
                    "Nodes": Nodes,
                    "nodes": nodes,
                    "Centroids": Centroids,
                    "centroids": centroids,
                }
            )
            data["results"].update(
                {
                    "Displacements": Displacements,
                    "displacements": displacements,
                }
            )
            # Assigning.
            if type(object) == gp.mesh.Mesh:
                object._Nodes = Nodes
                object._nodes = nodes
                object._Centroids = Centroids
                object._centroids = centroids
                object._Displacements = Displacements
                object._displacements = displacements
                object._calibrated = True
            if verbose:
                log.info("... Mesh Calibrated.")
        elif data["type"] == "Sequence":
            with alive_bar(
                len(data["meshes"]), bar="blocks", title="Calibrating Sequence ..."
            ) as bar:
                # Recursive.
                if data["file_settings"]["save_by_reference"] is True:
                    for i in range(len(data["meshes"])):
                        mesh = object._load_mesh(i, obj=True, verbose=False)
                        self.calibrate(object=mesh, verbose=False)
                        object._save_mesh(mesh, i, verbose=False)
                        bar()
                else:
                    for i in range(len(data["meshes"])):
                        self.calibrate(object=data["meshes"][i], verbose=False)
                        bar()
            # Storing.
            data["calibrated"] = True
            # Assigning.
            if type(object) == gp.sequence.Sequence:
                object._calibrated = True

    def _report(self, msg, error_type):
        if msg and error_type != "Warning":
            log.error(msg)
        elif msg and error_type == "Warning":
            log.warning(msg)
            return True
        if error_type == "ValueError" and msg:
            raise ValueError(msg)
        elif error_type == "TypeError" and msg:
            raise TypeError(msg)
        elif error_type == "IndexError" and msg:
            raise IndexError(msg)


class Calibration(CalibrationBase):
    def __init__(
        self,
        *,
        calibration_dir=".",
        common_name="",
        calibration_file_format=".jpg",
        dictionary=None,
        board_parameters=None,
        show=False,
        save=False,
    ):
        """
        Initialisation of geopyv calibration object.

        Parameters
        ----------
        calibration_dir : str, optional
            Directory of calibration images. Defaults to current working directory.
        common_name : str, optional
            If multiple image sets are within the folder, specifying the common name
            is necessary to identify which set of images to use.
        calibration_file_format : str, optional
                Image file type. Options are ".jpg", ".png" or ".bmp". Defaults to .jpg.

        """

        # Set initialised boolean.
        self.initialised = False

        # Check types.
        self._report(
            gp.check._check_type(calibration_dir, "calibration_dir", [str]), "TypeError"
        )
        if self._report(
            gp.check._check_path(calibration_dir, "calibration_dir"), "Warning"
        ):
            calibration_dir = gp.io._get_image_dir()
        calibration_dir = gp.check._check_character(calibration_dir, "/", -1)
        self._report(
            gp.check._check_type(
                calibration_file_format, "calibration_file_format", [str]
            ),
            "TypeError",
        )
        file_format = gp.check._check_character(calibration_file_format, ".", 0)
        self._report(
            gp.check._check_value(
                calibration_file_format,
                "calibration_file_format",
                [".jpg", ".png", ".bmp", ".tiff"],
            ),
            "ValueError",
        )

        # Store variables.
        self._calibration_dir = calibration_dir
        self._common_name = common_name
        self._file_format = file_format
        self._dictionary = dictionary
        self.solved = False
        self._unsolvable = False

        # Board Setup.
        board_settings = self._board_setup(board_parameters, show, save)

        # Load calibration images.
        try:
            self._calibration_images = glob.glob(
                self._common_name + "*" + self._file_format,
                root_dir=self._calibration_dir,
            )
            image_no = np.shape(self._calibration_images)[0]

            if image_no < 1:
                msg = "No images found in specified folder."
                log.error(msg)
                raise FileExistsError(msg)
            elif image_no < 10:
                msg = (
                    "{} images found. A minimum of 10 images is recommended."
                ).format(image_no)
                log.warning(msg)
            else:
                msg = ("{} images found.").format(image_no)
                log.info(msg)
        except Exception:
            log.error(
                ("Cannot find files with specified path: {}").format(
                    self._calibration_dir + self._common_name + "*" + self._file_format
                )
            )
            print(traceback.format_exc())
            return

        image = cv2.imread(
            self._calibration_dir + self._calibration_images[0], cv2.IMREAD_COLOR
        )
        self._image_size = np.shape(image)
        del image

        # Data.
        file_settings = {
            "calibration_dir": self._calibration_dir,
            "common_name": self._common_name,
            "calibration_images": self._calibration_images,
            "file_format": self._file_format,
            "image_size": self._image_size,
        }
        self.data = {
            "type": "Calibration",
            "solved": self.solved,
            "unsolvable": self._unsolvable,
            "file_settings": file_settings,
            "board_settings": board_settings,
        }

        self._initialised = True

    def _board_setup(self, board_parameters, show, save):
        """Private method for setting up the board."""

        # Board setup.
        columns, rows, square_length, marker_length = board_parameters
        self._board = aruco.CharucoBoard(
            (columns, rows), square_length, marker_length, self._dictionary
        )
        self._objpnts = self._board.getChessboardCorners()
        self._board_corners = self._board.getChessboardCorners()
        board_settings = {
            # "dictionary": self._dictionary,
            "columns": columns,
            "rows": rows,
            "corners": self._board_corners,
            "objpnts": self._objpnts,
            "square_length": square_length,
            "marker_length": marker_length,
        }

        if show:
            imboard = self._board.generateImage((290, 180))
            if save:
                cv2.imwrite(save + ".tiff", imboard)
            fig = plt.figure()
            ax = fig.add_subplot(1, 1, 1)
            plt.imshow(imboard, interpolation="nearest")
            ax.axis("off")
            plt.show()

        return board_settings

    def solve(self, *, index=None, binary=None, threshold=50):
        """
        Private method to calibrate the camera.
        """

        # Input checks.
        check = gp.check._check_type(binary, "binary", [float, int, type(None)])
        if check:
            try:
                binary = float(binary)
                self._report(gp.check._conversion(binary, "binary", float), "Warning")
            except Exception:
                self._report(check, "TypeError")
        self._report(
            gp.check._check_type(threshold, "threshold", [int, float]), "TypeError"
        )
        check = self._report(gp.check._check_type(index, "index", [int]), "Warning")
        if check:
            _ext_img = gp.io._load_img("extrinsic", load=False)
            index = int(re.findall(r"\d+", _ext_img)[-1])
        image_indices = np.asarray(
            [int(re.findall(r"\d+", x)[-1]) for x in self._calibration_images]
        )
        try:
            index = np.argwhere(image_indices == index)[0][0]
        except Exception:
            _ext_img = gp.io._load_img("extrinsic", load=False)
            index = int(re.findall(r"\d+", _ext_img)[-1])
            index = np.argwhere(image_indices == index)[0][0]

        # Store input.
        self._binary = binary
        self._index = index

        # Pre-process images for calibration.
        self._allCorners, self._allIds, self._imsize = self._read_chessboards(threshold)
        if len(self._allCorners) == 0:
            log.error("Unsolvable: no corners identified.")
            return
        self._imgpnts = []
        for i in range(len(self._allCorners)):
            self._imgpnts.append(self._allCorners[i].reshape(-1, 2))

        # Calibrate camera.
        self._calibrate_camera(self._allCorners, self._allIds, self._imsize)

        # Post-processing.
        self._extrinsic_matrix_generator()
        self._camera_angles()
        self._reprojection()

        # Store data.
        self.solved = True
        self.data.update(
            {
                "solved": self.solved,
                "intrinsic_matrix": self._intmat,
                "extrinsic_matrix": self._extmat,
                "distortion": self._dist,
                "camera_angles": [
                    self._alpha,
                    self._beta,
                    self._gamma,
                ],
                "projection": {
                    "imgpnts": self._imgpnts,
                    "objpnts": self._objpnts,
                    "reimgpnts": self._reimgpnts,
                },
                "calibration": {
                    "ret": self._ret,
                    "rotation": self._rot,
                    "translation": self._trans,
                    "corners": self._allCorners,
                    "ids": self._allIds,
                },
            }
        )

    def _read_chessboards(self, threshold):
        allCorners = []
        allIds = []
        acceptedImages = []

        # Sub-pixel detection criteria.
        criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 100, 0.00001)

        with alive_bar(
            np.shape(self._calibration_images)[0],
            dual_line=True,
            bar="blocks",
            title="Extracting...",
        ) as bar:
            for i in range(len(self._calibration_images)):
                bar.text = "Processing image {}...".format(
                    self._calibration_dir + self._calibration_images[i]
                )

                # Image loading.
                frame = cv2.imread(self._calibration_dir + self._calibration_images[i])
                frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
                frame = cv2.GaussianBlur(frame, (5, 5), 0)

                # Image binarisation (optional).
                if self._binary is not None:
                    ret, bin_frame = cv2.threshold(
                        frame, self._binary, 255, cv2.THRESH_BINARY
                    )

                    # Marker Detection.
                    arcnrs, arids, rejectedImgPoints = cv2.aruco.detectMarkers(
                        bin_frame,
                        self._dictionary,
                        parameters=cv2.aruco.DetectorParameters(),
                    )
                else:
                    # Marker Detection.
                    arcnrs, arids, rejectedImgPoints = cv2.aruco.detectMarkers(
                        frame,
                        self._dictionary,
                        parameters=cv2.aruco.DetectorParameters(),
                    )

                if len(arcnrs) > 0:
                    for arcnr in arcnrs:
                        # Sub-pixel refinement.
                        cv2.cornerSubPix(
                            frame,
                            arcnr,
                            winSize=(3, 3),
                            zeroZone=(-1, -1),
                            criteria=criteria,
                        )
                    # ChArUco interpolation.
                    ret, chcnrs, chids = cv2.aruco.interpolateCornersCharuco(
                        arcnrs, arids, frame, self._board
                    )
                    if chcnrs is not None and chids is not None and len(chcnrs) > 3:
                        if len(chcnrs) > threshold:
                            allCorners.append(chcnrs)
                            allIds.append(chids)
                            acceptedImages.append(self._calibration_images[i])
                bar()

        self._calibration_images = acceptedImages
        imsize = frame.shape
        return allCorners, allIds, imsize

    def _calibrate_camera(self, allCorners, allIds, imsize):
        """Private method to calibrate the camera."""

        log.info("Calibrating camera...")
        cameraMatrixInit = np.array(
            [
                [1000.0, 0.0, imsize[0] / 2.0],
                [0.0, 1000.0, imsize[1] / 2.0],
                [0.0, 0.0, 1.0],
            ]
        )
        distCoeffsInit = np.zeros((5, 1))
        (
            self._ret,
            self._intmat,
            _dist,
            _rot,
            _trans,
        ) = cv2.aruco.calibrateCameraCharuco(
            charucoCorners=allCorners,
            charucoIds=allIds,
            board=self._board,
            imageSize=imsize,
            cameraMatrix=cameraMatrixInit,
            distCoeffs=distCoeffsInit,
        )

        self._trans = np.asarray(_trans)
        self._rot = np.asarray(_rot)
        self._dist = _dist.flatten()

    def _reprojection(self):
        """Private method to generate reproject from object space
        to image space for the calibrationcoordinates for the purpose
        of error estimation."""

        self._reimgpnts = []
        for index in range(len(self._allCorners)):
            objpnts = self._find_objpnts(index)
            X_c = self._extmats[index] @ objpnts.T
            X_c /= X_c[2]
            r2 = X_c[0] ** 2 + X_c[1] ** 2
            f = (
                1
                + self._dist[0] * r2
                + self._dist[1] * r2**2
                + self._dist[4] * r2**3
            )
            X_pp = np.ones((np.shape(objpnts)[0], 3))
            X_pp[:, 0] = (
                X_c[0] * f
                + 2 * self._dist[2] * r2 * X_c[0] * X_c[1]
                + self._dist[3] * (r2 + 2 * X_c[0] ** 2)
            )
            X_pp[:, 1] = (
                X_c[1] * f
                + self._dist[2] * (r2 + 2 * X_c[1] ** 2)
                + 2 * self._dist[3] * r2 * X_c[0] * X_c[1]
            )
            self._reimgpnts.append(((self._intmat @ X_pp.T).T)[:, :2])

    def _find_objpnts(self, index):
        objpnts = np.ones((len(self._allIds[index]), 4))
        objpnts[:, :3] = self._objpnts[self._allIds[index]].flatten().reshape(-1, 3)
        return objpnts

    def _extrinsic_matrix_generator(self):
        """Private method for constructing the extrinsic matrices for calibration
        and the extrinsic matrix for mapping."""

        self._extmats = np.zeros((np.shape(self._rot)[0], 4, 4))
        self._extmats[:, :3, 3] = self._trans[:, :3].reshape(-1, 3)
        self._extmats[:, 3, 3] = 1
        for i in range(np.shape(self._extmats)[0]):
            self._extmats[i, :3, :3], _ = cv2.Rodrigues(self._rot[i])
        self._extmat = self._extmats[self._index]

    def _camera_angles(self):
        self._beta = -np.arcsin(self._extmat[2, 0])
        self._gamma = np.arcsin(self._extmat[2, 1] / np.cos(self._beta))
        self._alpha = np.arcsin(self._extmat[1, 0] / np.cos(self._beta))


class CalibrationResults(CalibrationBase):
    """
    Calibration Results class for geopyv.

    Parameters
    ----------
    data : dict
        geopyv data dict from Calibration object.


    Attributes
    ----------
    data : dict
        geopyv data dict from Calibration object.

    """

    def __init__(self, data):
        """Initialisation of geopyv CalibrationResults class."""
        self.data = data
        self._intmat = self.data["intrinsic_matrix"]
        self._extmat = self.data["extrinsic_matrix"]
        self._dist = self.data["distortion"]



"""

Speckle module for geopyv.

"""

import logging
import geopyv as gp
import numpy as np
import scipy as sp
from geopyv.object import Object
from PIL import Image as Im
from alive_progress import alive_bar

log = logging.getLogger(__name__)


class SpeckleBase(Object):
    """

    Speckle base class to be used as a mixin. Contains plot functionality.

    """

    def __init__(self):
        super().__init__(object_type="Speckle")
        """

        Speckle base class initialiser.

        """

    def _warp(self, i, ref, rot=False):
        _warp = np.zeros((np.shape(ref)[0], np.shape(self.data["comp"])[0]))
        delta = ref - self.data["origin"]
        _warp[:, 0] = (
            self.data["pm"][i][0]
            + self.data["pm"][i][2] * delta[:, 0]
            + self.data["pm"][i][4] * delta[:, 1]
            + 0.5 * self.data["pm"][i][6] * delta[:, 0] ** 2
            + self.data["pm"][i][8] * delta[:, 0] * delta[:, 1]
            + 0.5 * self.data["pm"][i][10] * delta[:, 1] ** 2
        )
        _warp[:, 1] = (
            self.data["pm"][i][1]
            + self.data["pm"][i][3] * delta[:, 0]
            + self.data["pm"][i][5] * delta[:, 1]
            + 0.5 * self.data["pm"][i][7] * delta[:, 0] ** 2
            + self.data["pm"][i][9] * delta[:, 0] * delta[:, 1]
            + 0.5 * self.data["pm"][i][11] * delta[:, 1] ** 2
        )
        _warp[:, 2] = (
            self.data["pm"][i][2]
            + self.data["pm"][i][6] * delta[:, 0]
            + self.data["pm"][i][8] * delta[:, 1]
        )
        _warp[:, 3] = (
            self.data["pm"][i][3]
            + self.data["pm"][i][7] * delta[:, 0]
            + self.data["pm"][i][9] * delta[:, 1]
        )
        _warp[:, 4] = (
            self.data["pm"][i][4]
            + self.data["pm"][i][8] * delta[:, 1]
            + self.data["pm"][i][10] * delta[:, 1]
        )
        _warp[:, 5] = (
            self.data["pm"][i][5]
            + self.data["pm"][i][9] * delta[:, 1]
            + self.data["pm"][i][11] * delta[:, 1]
        )
        _warp[:, 6:] = self.data["pm"][i][6:]

        if rot:
            _warp[:, :2] -= delta
        return _warp

    def _report(self, msg, error_type):
        if msg and error_type != "Warning":
            log.error(msg)
        elif msg and error_type == "Warning":
            log.warning(msg)
            return True
        if error_type == "ValueError" and msg:
            raise ValueError(msg)
        elif error_type == "TypeError" and msg:
            raise TypeError(msg)
        elif error_type == "IndexError" and msg:
            raise IndexError(msg)


class Speckle(SpeckleBase):
    """
    Speckle synthetic image generator.

    """

    def __init__(
        self,
        *,
        name=None,
        image_dir=None,
        file_format=".jpg",
        image_size_x=1001,
        image_size_y=1001,
        speckle_limits=1001,
        image_no=101,
        mmin=0,
        mmax=1,
        mtyp=0,
        comp=np.zeros(12),
        origin=np.asarray([501.0, 501.0]),
        noise=np.asarray([[0.0, 0.0], [0, 0]]),
        tmi=100,
        speckle_size=10,
    ):
        self._report(gp.check._check_type(name, "name", [str]), "Error")
        self._report(gp.check._check_type(image_dir, "image_dir", [str]), "Error")
        self._report(
            gp.check._check_type(file_format, "file_format", [str]), "TypeError"
        )
        file_format = gp.check._check_character(file_format, ".", 0)
        self._report(
            gp.check._check_value(
                file_format,
                "file_format",
                [".jpg", ".png", ".bmp", ".JPG", ".PNG", ".BMP"],
            ),
            "ValueError",
        )
        if self._report(gp.check._check_path(image_dir, "image_dir"), "Warning"):
            image_dir = gp.io._get_image_dir()
        image_dir = gp.check._check_character(image_dir, "/", -1)
        check = gp.check._check_type(image_no, "image_no", [int])
        if check:
            try:
                image_no = int(image_no)
                self._report(gp.check._conversion(image_no, "image_no", int), "Warning")
            except Exception:
                self._report(check, "TypeError")
        self._report(gp.check._check_range(image_no, "image_no", 1), "ValueError")
        self._report(gp.check._check_type(mmin, "mmin", [float, int]), "Error")
        self._report(gp.check._check_type(mmax, "mmax", [float, int]), "Error")
        self._report(gp.check._check_type(mtyp, "mtyp", [int]), "Error")
        self._report(gp.check._check_type(comp, "comp", [np.ndarray]), "Error")
        self._report(gp.check._check_axis(comp, "comp", 0, [12]), "Error")

        self._name = name
        self._image_dir = image_dir
        self._image_size_x = image_size_x
        self._image_size_y = image_size_y
        self._speckle_limits = speckle_limits
        self._file_format = file_format
        self._image_no = image_no
        self._mmin = mmin
        self._mmax = mmax
        self._mtyp = mtyp
        self._comp = comp
        self._origin = origin
        self._noise = noise
        self._tmi = tmi
        self._speckle_size = speckle_size
        self.solved = False
        self._ref_speckle = None

        # Data.
        self.data = {
            "type": "Speckle",
            "name": self._name,
            "solved": self.solved,
            "image_dir": self._image_dir,
            "file_format": self._file_format,
            "image_no": self._image_no,
            "comp": self._comp,
            "origin": self._origin,
            "noise": self._noise,
            "tmi": self._tmi,
            "speckle_size": self._speckle_size,
        }

    def solve(self, *, wrap=False, number=None, generate=True):
        self._wrap = wrap
        [self.X, self.Y] = np.meshgrid(
            range(self._image_size_x), range(self._image_size_y)
        )
        if generate:
            self._ref_speckle = self._speckle_distribution(number)
        self._mult()
        self.data.update(
            {"wrap": self._wrap, "ref_speckle": self._ref_speckle, "pm": self._pm}
        )
        self._image_generation()
        self.solved = True
        self.data["solved"] = True

    def _speckle_distribution(self, number):
        raw_grid = np.zeros((self._image_size_y, self._image_size_x))

        if number is not None:
            corner = np.asarray(
                [
                    0.5 * (self._image_size_x - self._speckle_limits),
                    0.5 * (self._image_size_y - self._speckle_limits),
                ]
            )
            speckles = corner + np.random.rand(number, 2) * self._speckle_limits
        else:
            i = 0
            mi = 0
            speckles = []
            with alive_bar(
                self._tmi,
                manual=True,
                dual_line=True,
                bar="blocks",
                title="Speckling...",
            ) as bar:
                while mi < self._tmi and i < 50000:
                    corner = np.asarray(
                        [
                            0.5 * self._image_size_y - self._speckle_limits,
                            0.5 * self._image_size_x - self._speckle_limits,
                        ]
                    )
                    new_speckle = corner + np.random.rand(2) * self._speckle_limits
                    speckles.append(new_speckle)
                    raw_grid += np.exp(
                        -(
                            (self.X - new_speckle[0]) ** 2
                            + (self.Y - new_speckle[1]) ** 2
                        )
                        / ((self._speckle_size**2) / 4)
                    )
                    final_grid = np.clip(raw_grid * 200, 0, 255)
                    mi = np.mean(final_grid)
                    i += 1
                    bar(min(mi / self._tmi, 1))
                    print(i)

        return np.asarray(speckles)

    def _mult(self):
        self._pm = np.zeros((self._image_no, np.shape(self._comp)[0]))
        self.noisem = np.zeros((self._image_no, 2))
        if self._mtyp == 0:
            self._mult = np.linspace(
                self._mmin, self._mmax, self._image_no, endpoint=True
            )
        elif self._mtyp == 1:
            self._mult = np.zeros((self._image_no))
            self._mult[1:] = np.logspace(
                np.log10(self._mmin),
                np.log10(self._mmax),
                self._image_no - 1,
                endpoint=True,
            )
        for i in range(self._image_no):
            self._pm[i] = self._comp * self._mult[i]
            self.noisem[i] = self._noise[0] * self._noise[1] * self._mult[i]

    def _image_generation(self):
        with alive_bar(
            self._image_no, dual_line=True, bar="blocks", title="Generating images..."
        ) as bar:
            for i in range(self._image_no):
                warp = self._warp(i, self._ref_speckle)
                warp[:, :2] += self._ref_speckle
                # _tar_speckle = (
                #     self._warp(i, self._ref_speckle)[:, :2] + self._ref_speckle
                # )
                # if self._wrap:
                #     _tar_speckle[:, 0] %= self._image_size_x
                #     _tar_speckle[:, 1] %= self._image_size_y
                # _grid = self._grid(i, _tar_speckle)
                _grid = self._grid(i, warp)
                self._create(i, _grid)
                bar()

    def _grid(self, i, warp):
        grid = np.zeros((self._image_size_y, self._image_size_x))
        if self.noisem[i, 0] != 0.0:
            warp[:, :2] = np.random.normal(loc=warp[:, :2], scale=self.noisem[i, 0])
        # covs = np.reshape(warp[:,2:6], (-1,2,2))*self._speckle_size**2
        for j in range(len(warp)):
            cov = np.identity(2)
            cov[0, 0] += warp[j, 2]
            cov[1, 1] += warp[j, 5]
            cov *= (self._speckle_size**2) / 8
            shear_matrix = np.asarray([[1.0, warp[j, 4]], [warp[j, 3], 1.0]])
            cov = np.dot(shear_matrix, np.dot(cov, shear_matrix.T))
            distribution = sp.stats.multivariate_normal(
                mean=warp[j, :2], cov=cov, allow_singular=True
            )

            a = max(int(warp[j, 1]) - 500, 0)
            b = min(int(warp[j, 1]) + 501, self._image_size_y)
            c = max(int(warp[j, 0]) - 100, 0)
            d = min(int(warp[j, 0]) + 101, self._image_size_x)
            positions = np.dstack((self.X[a:b, c:d], self.Y[a:b, c:d]))
            di = distribution.pdf(positions)
            di /= 1 / (2 * np.pi * np.sqrt(cov[0, 0]) * np.sqrt(cov[1, 1]))

            # 1/np.max(distribution.pdf(warp[j,:2]))
            # di = np.exp(
            #     -(
            #         (
            #             self.X[
            #                 a:b,
            #                 c:d,
            #             ]
            #             - _tar_speckle[j, 0]
            #         )
            #         ** 2
            #         + (
            #             self.Y[
            #                 a:b,
            #                 c:d,
            #             ]
            #             - _tar_speckle[j, 1]
            #         )
            #         ** 2
            #     )
            #     / ((self._speckle_size**2) / 4)
            # )
            # fig, ax = plt.subplots()
            # ax.contourf(self.X[a:b,c:d], self.Y[a:b,c:d], di)
            # plt.show()
            grid[
                a:b,
                c:d,
            ] += di
        if self.noisem[i, 1] != 0.0:
            grid = np.random.normal(loc=grid, scale=self.noisem[i, 1])
        grid = np.clip(grid * 200, 0, 255)
        return grid

    def _create(self, i, grid):
        """Method to store generated speckle pattern intensity matrix as a image file.

        Parameters
        ----------

        """
        im = Im.fromarray(np.uint8(grid))
        im.save(self._image_dir + self._name + "_" + str(i) + self._file_format)
        im.close()


class SpeckleResults(SpeckleBase):
    """

    SpeckleResults class for geopyv.

    Parameters
    ----------
    data : dict
        geopyv data dict from Speckle object.


    Attributes
    ----------
    data : dict
        geopyv data dict from Speckle object.

    """

    def __init__(self, data):
        """Initialisation of geopyv SpeckleResults class."""
        self.data = data



from skbuild import setup

setup(
    packages=["geopyv", "geopyv.geometry", "geopyv.gui", "geopyv.gui.selectors"],
    package_dir={"": "src"},
    cmake_install_dir="src/geopyv",
    include_package_data=True,
)
